import sys
import multiprocessing as mp
import traceback
import pickle

import numpy as np

from numba import cuda
from numba.cuda.cudadrv import drvapi, devicearray
from numba.cuda.testing import skip_on_cudasim, ContextResettingTestCase
from numba.tests.support import linux_only
import unittest

linux = sys.platform.startswith('linux')
has_mp_get_context = hasattr(mp, 'get_context')


def core_ipc_handle_test(the_work, result_queue):
    try:
        arr = the_work()
    except:
        # FAILED. propagate the exception as a string
        succ = False
        out = traceback.format_exc()
    else:
        # OK. send the ndarray back
        succ = True
        out = arr
    result_queue.put((succ, out))


def base_ipc_handle_test(handle, size, result_queue):
    def the_work():
        dtype = np.dtype(np.intp)
        with cuda.open_ipc_array(handle, shape=size // dtype.itemsize,
                                 dtype=dtype) as darr:
            # copy the data to host
            return darr.copy_to_host()

    core_ipc_handle_test(the_work, result_queue)


def serialize_ipc_handle_test(handle, result_queue):
    def the_work():
        dtype = np.dtype(np.intp)
        darr = handle.open_array(cuda.current_context(),
                                 shape=handle.size // dtype.itemsize,
                                 dtype=dtype)
        # copy the data to host
        arr = darr.copy_to_host()
        handle.close()
        return arr

    core_ipc_handle_test(the_work, result_queue)


def ipc_array_test(ipcarr, result_queue):
    try:
        with ipcarr as darr:
            arr = darr.copy_to_host()
            try:
                # should fail to reopen
                with ipcarr:
                    pass
            except ValueError as e:
                if str(e) != 'IpcHandle is already opened':
                    raise AssertionError('invalid exception message')
            else:
                raise AssertionError('did not raise on reopen')

    except:
        # FAILED. propagate the exception as a string
        succ = False
        out = traceback.format_exc()
    else:
        # OK. send the ndarray back
        succ = True
        out = arr
    result_queue.put((succ, out))


@linux_only
@unittest.skipUnless(has_mp_get_context, "requires multiprocessing.get_context")
@skip_on_cudasim('Ipc not available in CUDASIM')
class TestIpcMemory(ContextResettingTestCase):
    def test_ipc_handle(self):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)

        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)

        # manually prepare for serialization as bytes
        handle_bytes = bytes(ipch.handle)
        size = ipch.size

        # spawn new process for testing
        ctx = mp.get_context('spawn')
        result_queue = ctx.Queue()
        args = (handle_bytes, size, result_queue)
        proc = ctx.Process(target=base_ipc_handle_test, args=args)
        proc.start()
        succ, out = result_queue.get()
        if not succ:
            self.fail(out)
        else:
            np.testing.assert_equal(arr, out)
        proc.join(3)

    def check_ipc_handle_serialization(self, index_arg=None):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        if index_arg is not None:
            devarr = devarr[index_arg]
        expect = devarr.copy_to_host()

        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)

        # pickle
        buf = pickle.dumps(ipch)
        ipch_recon = pickle.loads(buf)
        self.assertIs(ipch_recon.base, None)
        self.assertEqual(tuple(ipch_recon.handle), tuple(ipch.handle))
        self.assertEqual(ipch_recon.size, ipch.size)

        # spawn new process for testing
        ctx = mp.get_context('spawn')
        result_queue = ctx.Queue()
        args = (ipch, result_queue)
        proc = ctx.Process(target=serialize_ipc_handle_test, args=args)
        proc.start()
        succ, out = result_queue.get()
        if not succ:
            self.fail(out)
        else:
            np.testing.assert_equal(expect, out)
        proc.join(3)

    def test_ipc_handle_serialization(self):
        # test no slicing
        self.check_ipc_handle_serialization()
        # slicing tests
        self.check_ipc_handle_serialization(slice(3, None))
        self.check_ipc_handle_serialization(slice(3, 8))
        self.check_ipc_handle_serialization(slice(None, 8))

    def check_ipc_array(self, index_arg=None):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        # Slice
        if index_arg is not None:
            devarr = devarr[index_arg]
        expect = devarr.copy_to_host()
        ipch = devarr.get_ipc_handle()

        # spawn new process for testing
        ctx = mp.get_context('spawn')
        result_queue = ctx.Queue()
        args = (ipch, result_queue)
        proc = ctx.Process(target=ipc_array_test, args=args)
        proc.start()
        succ, out = result_queue.get()
        if not succ:
            self.fail(out)
        else:
            np.testing.assert_equal(expect, out)
        proc.join(3)

    def test_ipc_array(self):
        # test no slicing
        self.check_ipc_array()
        # slicing tests
        self.check_ipc_array(slice(3, None))
        self.check_ipc_array(slice(3, 8))
        self.check_ipc_array(slice(None, 8))

@unittest.skipIf(linux, 'Only on OS other than Linux')
@skip_on_cudasim('Ipc not available in CUDASIM')
class TestIpcNotSupported(ContextResettingTestCase):
    def test_unsupported(self):
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)
        with self.assertRaises(OSError) as raises:
            devarr.get_ipc_handle()
        errmsg = str(raises.exception)
        self.assertIn('OS does not support CUDA IPC', errmsg)


def staged_ipc_handle_test(handle, device_num, result_queue):
    def the_work():
        with cuda.gpus[device_num]:
            this_ctx = cuda.devices.get_context()
            deviceptr = handle.open_staged(this_ctx)
            arrsize = handle.size // np.dtype(np.intp).itemsize
            hostarray = np.zeros(arrsize, dtype=np.intp)
            cuda.driver.device_to_host(
                hostarray, deviceptr,  size=handle.size,
                )
            handle.close()
        return hostarray

    core_ipc_handle_test(the_work, result_queue)


def staged_ipc_array_test(ipcarr, device_num, result_queue):
    try:
        with cuda.gpus[device_num]:
            this_ctx = cuda.devices.get_context()
            with ipcarr as darr:
                arr = darr.copy_to_host()
                try:
                    # should fail to reopen
                    with ipcarr:
                        pass
                except ValueError as e:
                    if str(e) != 'IpcHandle is already opened':
                        raise AssertionError('invalid exception message')
                else:
                    raise AssertionError('did not raise on reopen')
    except:
        # FAILED. propagate the exception as a string
        succ = False
        out = traceback.format_exc()
    else:
        # OK. send the ndarray back
        succ = True
        out = arr
    result_queue.put((succ, out))


@linux_only
@unittest.skipUnless(has_mp_get_context, "requires multiprocessing.get_context")
@skip_on_cudasim('Ipc not available in CUDASIM')
class TestIpcStaged(ContextResettingTestCase):
    def test_staged(self):
        # prepare data for IPC
        arr = np.arange(10, dtype=np.intp)
        devarr = cuda.to_device(arr)

        # spawn new process for testing
        mpctx = mp.get_context('spawn')
        result_queue = mpctx.Queue()

        # create IPC handle
        ctx = cuda.current_context()
        ipch = ctx.get_ipc_handle(devarr.gpu_data)
        # pickle
        buf = pickle.dumps(ipch)
        ipch_recon = pickle.loads(buf)
        self.assertIs(ipch_recon.base, None)
        self.assertEqual(tuple(ipch_recon.handle), tuple(ipch.handle))
        self.assertEqual(ipch_recon.size, ipch.size)

        # Test on every CUDA devices
        for device_num in range(len(cuda.gpus)):
            args = (ipch, device_num, result_queue)
            proc = mpctx.Process(target=staged_ipc_handle_test, args=args)
            proc.start()
            succ, out = result_queue.get()
            proc.join(3)
            if not succ:
                self.fail(out)
            else:
                np.testing.assert_equal(arr, out)

    def test_ipc_array(self):
        for device_num in range(len(cuda.gpus)):
            # prepare data for IPC
            arr = np.random.random(10)
            devarr = cuda.to_device(arr)
            ipch = devarr.get_ipc_handle()

            # spawn new process for testing
            ctx = mp.get_context('spawn')
            result_queue = ctx.Queue()
            args = (ipch, device_num, result_queue)
            proc = ctx.Process(target=staged_ipc_array_test, args=args)
            proc.start()
            succ, out = result_queue.get()
            proc.join(3)
            if not succ:
                self.fail(out)
            else:
                np.testing.assert_equal(arr, out)


if __name__ == '__main__':
    unittest.main()



"""
Implement background services for the application.
This is implemented as a cooperative concurrent task.
"""

import functools


class Service(object):
    def __init__(self, name="unnamed", arg=None):
        self.name = name
        self.enabled = True
        self.arg = arg
        self._task = self.process(self.arg)
        next(self._task)

    def service(self):
        """
        Request for the service task.
        Servicing is disabled if it is disabled thourght the "enabled"
        attribute.  When the task is executing, the service is disabled to
        avoid recursion.
        """
        if self.enabled:
            enable = self.enabled
            try:
                # Prevent recursion
                self.enabled = False
                next(self._task)
            finally:
                self.enabled = enable

    def process(self, arg):
        """
        Overrided to implement the service task.
        This must be a generator.
        Use `yield` to return control.
        """
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.service()

    def after(self, fn):
        """
        A decorator for a function. Service is triggered on return.
        """
        @functools.wraps(fn)
        def wrap(*args, **kws):
            with self:
                return fn(*args, **kws)
        return wrap

# -----------------------------------------------------------------------------
# The rest are for testing


class HelloService(Service):
    def process(self, arg):
        count = 0
        yield
        while True:
            print("Hello", count)
            count += 1
            yield

def test():
    serv = HelloService("my.hello")
    print("1")
    serv.service()
    print("2")
    serv.service()

    with serv:
        print("3")

    @serv.after
    def nested():
        print("4")

    nested()


if __name__ == '__main__':
    test()



"""
This file will be copied to a temporary directory in order to
exercise caching compiled Numba functions.

See test_dispatcher.py.
"""

import sys

import numpy as np

from numba import jit, generated_jit, prange
from numba.core import types

from numba.tests.ctypes_usecases import c_sin
from numba.tests.support import TestCase, captured_stderr


@jit(cache=True, nopython=True)
def simple_usecase(x):
    return x

def simple_usecase_caller(x):
    return simple_usecase(x)


@jit(cache=True, nopython=True)
def add_usecase(x, y):
    return x + y + Z


@jit(cache=True, forceobj=True)
def add_objmode_usecase(x, y):
    object()
    return x + y + Z


@jit(nopython=True)
def add_nocache_usecase(x, y):
    return x + y + Z


@generated_jit(cache=True, nopython=True)
def generated_usecase(x, y):
    if isinstance(x, types.Complex):
        def impl(x, y):
            return x + y
    else:
        def impl(x, y):
            return x - y
    return impl


@jit(cache=True, nopython=True)
def inner(x, y):
    return x + y + Z

@jit(cache=True, nopython=True)
def outer(x, y):
    return inner(-y, x)

@jit(cache=False, nopython=True)
def outer_uncached(x, y):
    return inner(-y, x)


@jit(cache=True, forceobj=True)
def looplifted(n):
    object()
    res = 0
    for i in range(n):
        res = res + i
    return res


@jit(cache=True, nopython=True)
def use_c_sin(x):
    return c_sin(x)

@jit(cache=True, nopython=True)
def use_c_sin_nest1(x):
    return use_c_sin(x)

@jit(cache=True, nopython=True)
def use_c_sin_nest2(x):
    return use_c_sin_nest1(x)


@jit(cache=True, nopython=True)
def ambiguous_function(x):
    return x + 2

renamed_function1 = ambiguous_function

@jit(cache=True, nopython=True)
def ambiguous_function(x):
    return x + 6

renamed_function2 = ambiguous_function


def make_closure(x):
    @jit(cache=True, nopython=True)
    def closure(y):
        return x + y

    return closure

closure1 = make_closure(3)
closure2 = make_closure(5)


biggie = np.arange(10**6)

@jit(cache=True, nopython=True)
def use_big_array():
    return biggie


Z = 1

# Exercise returning a record instance.  This used to hardcode the dtype
# pointer's value in the bitcode.

packed_record_type = np.dtype([('a', np.int8), ('b', np.float64)])
aligned_record_type = np.dtype([('a', np.int8), ('b', np.float64)], align=True)

packed_arr = np.empty(2, dtype=packed_record_type)
for i in range(packed_arr.size):
    packed_arr[i]['a'] = i + 1
    packed_arr[i]['b'] = i + 42.5

aligned_arr = np.array(packed_arr, dtype=aligned_record_type)

@jit(cache=True, nopython=True)
def record_return(ary, i):
    return ary[i]


class _TestModule(TestCase):
    """
    Tests for functionality of this module's functions.
    Note this does not define any "test_*" method, instead check_module()
    should be called by hand.
    """

    def check_module(self, mod):
        self.assertPreciseEqual(mod.add_usecase(2, 3), 6)
        self.assertPreciseEqual(mod.add_objmode_usecase(2, 3), 6)
        self.assertPreciseEqual(mod.outer_uncached(3, 2), 2)
        self.assertPreciseEqual(mod.outer(3, 2), 2)
        self.assertPreciseEqual(mod.generated_usecase(3, 2), 1)

        packed_rec = mod.record_return(mod.packed_arr, 1)
        self.assertPreciseEqual(tuple(packed_rec), (2, 43.5))
        aligned_rec = mod.record_return(mod.aligned_arr, 1)
        self.assertPreciseEqual(tuple(aligned_rec), (2, 43.5))

    # For 2.x
    def runTest(self):
        raise NotImplementedError


def self_test():
    mod = sys.modules[__name__]
    _TestModule().check_module(mod)


@jit(parallel=True, cache=True, nopython=True)
def parfor_usecase(ary):
    return ary * ary + ary



from collections import namedtuple
import contextlib
import itertools
import math
import sys
import ctypes as ct
import numpy as np

from numba.core.compiler import compile_isolated, Flags
from numba import jit, typeof
import unittest
from numba import testing
from numba.core import types, utils, errors
from numba.tests.support import TestCase, MemoryLeakMixin, tag
from numba.experimental import jitclass


enable_pyobj_flags = Flags()
enable_pyobj_flags.set("enable_pyobject")

force_pyobj_flags = Flags()
force_pyobj_flags.set("force_pyobject")

Point = namedtuple('Point', ('a', 'b'))


def noop(x):
    pass

def unbox_usecase(x):
    """
    Expect a list of numbers
    """
    res = 0
    for v in x:
        res += v
    return res

def unbox_usecase2(x):
    """
    Expect a list of tuples
    """
    res = 0
    for v in x:
        res += len(v)
    return res

def unbox_usecase3(x):
    """
    Expect a (number, list of numbers) tuple.
    """
    a, b = x
    res = a
    for v in b:
        res += v
    return res

def unbox_usecase4(x):
    """
    Expect a (number, list of tuples) tuple.
    """
    a, b = x
    res = a
    for v in b:
        res += len(v)
    return res


def create_list(x, y, z):
    return [x, y, z]

def create_nested_list(x, y, z, a, b, c):
    return [[x, y, z], [a, b, c]]

def list_comprehension1():
    return sum([x**2 for x in range(10)])

def list_comprehension2():
    return sum([x for x in range(10) if x % 2 == 0])

def list_comprehension3():
    return sum([math.pow(x, 2) for x in range(10)])

def list_comprehension4():
    return sum([x * y for x in range(10) for y in range(10)])

def list_comprehension5():
    return [x * 2 for x in range(10)]

def list_comprehension6():
    return [[x for x in range(y)] for y in range(3)]


def list_constructor(n):
    return list(range(n))

def list_constructor_empty():
    # cannot be typed, list is empty and no typing information is present to
    # infer a type
    return list()

def list_constructor_empty_but_typeable(n):
    # can be typed, list is empty but later append has typing info that allows
    # for inference
    y = list()
    return y.append(n)

def list_append(n):
    l = []
    l.append(42)
    for i in range(n):
        l.append(i)
    return l

def list_append_heterogeneous(n):
    l = []
    l.append(42.0)
    for i in range(n):
        l.append(i)
    return l

def list_extend(n):
    l = []
    # A non-list iterable and a list
    l.extend(range(n))
    l.extend(l[:-1])
    l.extend(range(n, 0, -1))
    return l

def list_extend_heterogeneous(n):
    l = []
    # Extend with various iterables, including lists, with different types
    l.extend(range(n))
    l.extend(l[:-1])
    l.extend((5, 42))
    l.extend([123.0])
    return l

def list_pop0(n):
    l = list(range(n))
    res = 0
    while len(l) > 0:
        res += len(l) * l.pop()
    return res

def list_pop1(n, i):
    l = list(range(n))
    x = l.pop(i)
    return x, l

def list_len(n):
    l = list(range(n))
    return len(l)

def list_getitem(n):
    l = list(range(n))
    res = 0
    # Positive indices
    for i in range(len(l)):
        res += i * l[i]
    # Negative indices
    for i in range(-len(l), 0):
        res -= i * l[i]
    return res

def list_setitem(n):
    l = list(range(n))
    res = 0
    # Positive indices
    for i in range(len(l)):
        l[i] = i * l[i]
    # Negative indices
    for i in range(-len(l), 0):
        l[i] = i * l[i]
    for i in range(len(l)):
        res += l[i]
    return res

def list_getslice2(n, start, stop):
    l = list(range(n))
    return l[start:stop]

def list_getslice3(n, start, stop, step):
    l = list(range(n))
    return l[start:stop:step]

def list_setslice2(n, n_source, start, stop):
    # Generic setslice with size change
    l = list(range(n))
    v = list(range(100, 100 + n_source))
    l[start:stop] = v
    return l

def list_setslice3(n, start, stop, step):
    l = list(range(n))
    v = l[start:stop:step]
    for i in range(len(v)):
        v[i] += 100
    l[start:stop:step] = v
    return l

def list_setslice3_arbitrary(n, n_src, start, stop, step):
    l = list(range(n))
    l[start:stop:step] = list(range(100, 100 + n_src))
    return l

def list_delslice0(n):
    l = list(range(n))
    del l[:]
    return l

def list_delslice1(n, start, stop):
    l = list(range(n))
    del l[start:]
    del l[:stop]
    return l

def list_delslice2(n, start, stop):
    l = list(range(n))
    del l[start:stop]
    return l

def list_clear(n):
    l = list(range(n))
    l.clear()
    return l

def list_copy(n):
    l = list(range(n))
    ll = l.copy()
    l.append(42)
    return l, ll

def list_iteration(n):
    l = list(range(n))
    res = 0
    for i, v in enumerate(l):
        res += i * v
    return res

def list_contains(n):
    l = list(range(n))
    return (0 in l, 1 in l, n - 1 in l, n in l,
            0 not in l, 1 not in l, n - 1 not in l, n not in l,
            )

def list_index1(n, v):
    l = list(range(n, 0, -1))
    return l.index(v)

def list_index2(n, v, start):
    l = list(range(n, 0, -1))
    return l.index(v, start)

def list_index3(n, v, start, stop):
    l = list(range(n, 0, -1))
    return l.index(v, start, stop)

def list_remove(n, v):
    l = list(range(n - 1, -1, -1))
    l.remove(v)
    return l

def list_insert(n, pos, v):
    l = list(range(0, n))
    l.insert(pos, v)
    return l

def list_count(n, v):
    l = []
    for x in range(n):
        l.append(x & 3)
    return l.count(v)

def list_reverse(n):
    l = list(range(n))
    l.reverse()
    return l

def list_add(m, n):
    a = list(range(0, m))
    b = list(range(100, 100 + n))
    res = a + b
    res.append(42)   # check result is a copy
    return a, b, res

def list_add_heterogeneous():
    a = [1]
    b = [2.0]
    c = a + b
    d = b + a
    # check result is a copy
    a.append(3)
    b.append(4.0)
    return a, b, c, d

def list_add_inplace(m, n):
    a = list(range(0, m))
    b = list(range(100, 100 + n))
    a += b
    return a, b

def list_add_inplace_heterogeneous():
    a = [1]
    b = [2.0]
    a += b
    b += a
    return a, b

def list_mul(n, v):
    a = list(range(n))
    return a * v

def list_mul_inplace(n, v):
    a = list(range(n))
    a *= v
    return a

def list_bool(n):
    a = list(range(n))
    return bool(a), (True if a else False)

def eq_usecase(a, b):
    return list(a) == list(b)

def ne_usecase(a, b):
    return list(a) != list(b)

def gt_usecase(a, b):
    return list(a) > list(b)

def ge_usecase(a, b):
    return list(a) >= list(b)

def lt_usecase(a, b):
    return list(a) < list(b)

def le_usecase(a, b):
    return list(a) <= list(b)

def identity_usecase(n):
    a = list(range(n))
    b = a
    c = a[:]
    return (a is b), (a is not b), (a is c), (a is not c)

def bool_list_usecase():
    # Exercise getitem, setitem, iteration with bool values (issue #1373)
    l = [False]
    l[0] = True
    x = False
    for v in l:
        x = x ^ v
    return l, x

def reflect_simple(l, ll):
    x = l.pop()
    y = l.pop()
    l[0] = 42.
    l.extend(ll)
    return l, x, y

def reflect_conditional(l, ll):
    # `l` may or may not actually reflect a Python list
    if ll[0]:
        l = [11., 22., 33., 44.]
    x = l.pop()
    y = l.pop()
    l[0] = 42.
    l.extend(ll)
    return l, x, y

def reflect_exception(l):
    l.append(42)
    raise ZeroDivisionError

def reflect_dual(l, ll):
    l.append(ll.pop())
    return l is ll


class TestLists(MemoryLeakMixin, TestCase):

    def test_create_list(self):
        pyfunc = create_list
        cr = compile_isolated(pyfunc, (types.int32, types.int32, types.int32))
        cfunc = cr.entry_point
        self.assertEqual(cfunc(1, 2, 3), pyfunc(1, 2, 3))

    def test_create_nested_list(self):
        pyfunc = create_nested_list
        cr = compile_isolated(pyfunc, (types.int32, types.int32, types.int32,
                                       types.int32, types.int32, types.int32))
        cfunc = cr.entry_point
        self.assertEqual(cfunc(1, 2, 3, 4, 5, 6), pyfunc(1, 2, 3, 4, 5, 6))

    @testing.allow_interpreter_mode
    def test_list_comprehension(self):
        list_tests = [list_comprehension1,
                      list_comprehension2,
                      list_comprehension3,
                      list_comprehension4,
                      list_comprehension5,
                      list_comprehension6]

        for test in list_tests:
            pyfunc = test
            cr = compile_isolated(pyfunc, ())
            cfunc = cr.entry_point
            self.assertEqual(cfunc(), pyfunc())

    def check_unary_with_size(self, pyfunc, precise=True):
        cfunc = jit(nopython=True)(pyfunc)
        # Use various sizes, to stress the allocation algorithm
        for n in [0, 3, 16, 70, 400]:
            eq = self.assertPreciseEqual if precise else self.assertEqual
            eq(cfunc(n), pyfunc(n))

    def test_constructor(self):
        self.check_unary_with_size(list_constructor)

    def test_constructor_empty(self):
        self.disable_leak_check()
        cfunc = jit(nopython=True)(list_constructor_empty)
        with self.assertRaises(errors.TypingError) as raises:
            cfunc()
        errmsg = str(raises.exception)
        self.assertIn("Cannot infer the type of variable", errmsg)
        self.assertIn("list(undefined)", errmsg)
        # check error message went in
        self.assertIn("For Numba to be able to compile a list", errmsg)

    def test_constructor_empty_but_typeable(self):
        args = [np.int32(1), 10., 1 + 3j, [7], [17., 14.], np.array([10])]
        pyfunc = list_constructor_empty_but_typeable
        for arg in args:
            cfunc = jit(nopython=True)(pyfunc)
            expected = pyfunc(arg)
            got = cfunc(arg)
            self.assertPreciseEqual(got, expected)

    def test_append(self):
        self.check_unary_with_size(list_append)

    def test_append_heterogeneous(self):
        self.check_unary_with_size(list_append_heterogeneous, precise=False)

    def test_extend(self):
        self.check_unary_with_size(list_extend)

    def test_extend_heterogeneous(self):
        self.check_unary_with_size(list_extend_heterogeneous, precise=False)

    def test_pop0(self):
        self.check_unary_with_size(list_pop0)

    def test_pop1(self):
        pyfunc = list_pop1
        cfunc = jit(nopython=True)(pyfunc)
        for n in [5, 40]:
            for i in [0, 1, n - 2, n - 1, -1, -2, -n + 3, -n + 1]:
                expected = pyfunc(n, i)
                self.assertPreciseEqual(cfunc(n, i), expected)

    def test_pop_errors(self):
        # XXX References are leaked when an exception is raised
        self.disable_leak_check()
        cfunc = jit(nopython=True)(list_pop1)
        with self.assertRaises(IndexError) as cm:
            cfunc(0, 5)
        self.assertEqual(str(cm.exception), "pop from empty list")
        with self.assertRaises(IndexError) as cm:
            cfunc(1, 5)
        self.assertEqual(str(cm.exception), "pop index out of range")

    def test_insert(self):
        pyfunc = list_insert
        cfunc = jit(nopython=True)(pyfunc)
        for n in [5, 40]:
            indices = [0, 1, n - 2, n - 1, n + 1, -1, -2, -n + 3, -n - 1]
            for i in indices:
                expected = pyfunc(n, i, 42)
                self.assertPreciseEqual(cfunc(n, i, 42), expected)

    def test_len(self):
        self.check_unary_with_size(list_len)

    def test_getitem(self):
        self.check_unary_with_size(list_getitem)

    def test_setitem(self):
        self.check_unary_with_size(list_setitem)

    def check_slicing2(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        sizes = [5, 40]
        for n in sizes:
            indices = [0, 1, n - 2, -1, -2, -n + 3, -n - 1, -n]
            for start, stop in itertools.product(indices, indices):
                expected = pyfunc(n, start, stop)
                self.assertPreciseEqual(cfunc(n, start, stop), expected)

    def test_getslice2(self):
        self.check_slicing2(list_getslice2)

    def test_setslice2(self):
        pyfunc = list_setslice2
        cfunc = jit(nopython=True)(pyfunc)
        sizes = [5, 40]
        for n, n_src in itertools.product(sizes, sizes):
            indices = [0, 1, n - 2, -1, -2, -n + 3, -n - 1, -n]
            for start, stop in itertools.product(indices, indices):
                expected = pyfunc(n, n_src, start, stop)
                self.assertPreciseEqual(cfunc(n, n_src, start, stop), expected)

    def test_getslice3(self):
        pyfunc = list_getslice3
        cfunc = jit(nopython=True)(pyfunc)
        for n in [10]:
            indices = [0, 1, n - 2, -1, -2, -n + 3, -n - 1, -n]
            steps = [4, 1, -1, 2, -3]
            for start, stop, step in itertools.product(indices, indices, steps):
                expected = pyfunc(n, start, stop, step)
                self.assertPreciseEqual(cfunc(n, start, stop, step), expected)

    def test_setslice3(self):
        pyfunc = list_setslice3
        cfunc = jit(nopython=True)(pyfunc)
        for n in [10]:
            indices = [0, 1, n - 2, -1, -2, -n + 3, -n - 1, -n]
            steps = [4, 1, -1, 2, -3]
            for start, stop, step in itertools.product(indices, indices, steps):
                expected = pyfunc(n, start, stop, step)
                self.assertPreciseEqual(cfunc(n, start, stop, step), expected)

    def test_setslice3_resize(self):
        # XXX References are leaked when an exception is raised
        self.disable_leak_check()
        pyfunc = list_setslice3_arbitrary
        cfunc = jit(nopython=True)(pyfunc)
        # step == 1 => can resize
        cfunc(5, 10, 0, 2, 1)
        # step != 1 => cannot resize
        with self.assertRaises(ValueError) as cm:
            cfunc(5, 100, 0, 3, 2)
        self.assertIn("cannot resize", str(cm.exception))

    def test_delslice0(self):
        self.check_unary_with_size(list_delslice0)

    def test_delslice1(self):
        self.check_slicing2(list_delslice1)

    def test_delslice2(self):
        self.check_slicing2(list_delslice2)

    def test_invalid_slice(self):
        self.disable_leak_check()
        pyfunc = list_getslice3
        cfunc = jit(nopython=True)(pyfunc)
        with self.assertRaises(ValueError) as cm:
            cfunc(10, 1, 2, 0)
        self.assertEqual(str(cm.exception), "slice step cannot be zero")

    def test_iteration(self):
        self.check_unary_with_size(list_iteration)

    def test_reverse(self):
        self.check_unary_with_size(list_reverse)

    def test_contains(self):
        self.check_unary_with_size(list_contains)

    def check_index_result(self, pyfunc, cfunc, args):
        try:
            expected = pyfunc(*args)
        except ValueError:
            with self.assertRaises(ValueError):
                cfunc(*args)
        else:
            self.assertPreciseEqual(cfunc(*args), expected)

    def test_index1(self):
        self.disable_leak_check()
        pyfunc = list_index1
        cfunc = jit(nopython=True)(pyfunc)
        for v in (0, 1, 5, 10, 99999999):
            self.check_index_result(pyfunc, cfunc, (16, v))

    def test_index2(self):
        self.disable_leak_check()
        pyfunc = list_index2
        cfunc = jit(nopython=True)(pyfunc)
        n = 16
        for v in (0, 1, 5, 10, 99999999):
            indices = [0, 1, n - 2, n - 1, n + 1, -1, -2, -n + 3, -n - 1]
            for start in indices:
                self.check_index_result(pyfunc, cfunc, (16, v, start))

    def test_index3(self):
        self.disable_leak_check()
        pyfunc = list_index3
        cfunc = jit(nopython=True)(pyfunc)
        n = 16
        for v in (0, 1, 5, 10, 99999999):
            indices = [0, 1, n - 2, n - 1, n + 1, -1, -2, -n + 3, -n - 1]
            for start, stop in itertools.product(indices, indices):
                self.check_index_result(pyfunc, cfunc, (16, v, start, stop))

    def test_remove(self):
        pyfunc = list_remove
        cfunc = jit(nopython=True)(pyfunc)
        n = 16
        for v in (0, 1, 5, 15):
            expected = pyfunc(n, v)
            self.assertPreciseEqual(cfunc(n, v), expected)

    def test_remove_error(self):
        self.disable_leak_check()
        pyfunc = list_remove
        cfunc = jit(nopython=True)(pyfunc)
        with self.assertRaises(ValueError) as cm:
            cfunc(10, 42)
        self.assertEqual(str(cm.exception), "list.remove(x): x not in list")

    def test_count(self):
        pyfunc = list_count
        cfunc = jit(nopython=True)(pyfunc)
        for v in range(5):
            self.assertPreciseEqual(cfunc(18, v), pyfunc(18, v))

    def test_clear(self):
        self.check_unary_with_size(list_clear)

    def test_copy(self):
        self.check_unary_with_size(list_copy)

    def check_add(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        sizes = [0, 3, 50, 300]
        for m, n in itertools.product(sizes, sizes):
            expected = pyfunc(m, n)
            self.assertPreciseEqual(cfunc(m, n), expected)

    def test_add(self):
        self.check_add(list_add)

    def test_add_heterogeneous(self):
        pyfunc = list_add_heterogeneous
        cfunc = jit(nopython=True)(pyfunc)
        expected = pyfunc()
        self.assertEqual(cfunc(), expected)

    def test_add_inplace(self):
        self.check_add(list_add_inplace)

    def test_add_inplace_heterogeneous(self):
        pyfunc = list_add_inplace_heterogeneous
        cfunc = jit(nopython=True)(pyfunc)
        expected = pyfunc()
        self.assertEqual(cfunc(), expected)

    def check_mul(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        for n in [0, 3, 50, 300]:
            for v in [1, 2, 3, 0, -1, -42]:
                expected = pyfunc(n, v)
                self.assertPreciseEqual(cfunc(n, v), expected)

    def test_mul(self):
        self.check_mul(list_mul)

    def test_mul_inplace(self):
        self.check_mul(list_mul_inplace)

    @unittest.skipUnless(sys.maxsize >= 2**32,
                         "need a 64-bit system to test for MemoryError")
    def test_mul_error(self):
        self.disable_leak_check()
        pyfunc = list_mul
        cfunc = jit(nopython=True)(pyfunc)
        # Fail in malloc()
        with self.assertRaises(MemoryError):
            cfunc(1, 2**58)
        if sys.platform.startswith('darwin'):
            libc = ct.CDLL('libc.dylib')
            libc.printf("###Please ignore the above error message i.e. \
can't allocate region. It is in fact the purpose of this test to \
request more memory than can be provided###\n".encode("UTF-8"))
        # Overflow size computation when multiplying by item size
        with self.assertRaises(MemoryError):
            cfunc(1, 2**62)

    def test_bool(self):
        pyfunc = list_bool
        cfunc = jit(nopython=True)(pyfunc)
        for n in [0, 1, 3]:
            expected = pyfunc(n)
            self.assertPreciseEqual(cfunc(n), expected)

    def test_list_passing(self):
        # Check one can pass a list from a Numba function to another
        @jit(nopython=True)
        def inner(lst):
            return len(lst), lst[-1]

        @jit(nopython=True)
        def outer(n):
            l = list(range(n))
            return inner(l)

        self.assertPreciseEqual(outer(5), (5, 4))

    def _test_compare(self, pyfunc):
        def eq(args):
            self.assertIs(cfunc(*args), pyfunc(*args),
                          "mismatch for arguments %s" % (args,))

        cfunc = jit(nopython=True)(pyfunc)
        eq(((1, 2), (1, 2)))
        eq(((1, 2, 3), (1, 2)))
        eq(((1, 2), (1, 2, 3)))
        eq(((1, 2, 4), (1, 2, 3)))
        eq(((1.0, 2.0, 3.0), (1, 2, 3)))
        eq(((1.0, 2.0, 3.5), (1, 2, 3)))

    def test_eq(self):
        self._test_compare(eq_usecase)

    def test_ne(self):
        self._test_compare(ne_usecase)

    def test_le(self):
        self._test_compare(le_usecase)

    def test_lt(self):
        self._test_compare(lt_usecase)

    def test_ge(self):
        self._test_compare(ge_usecase)

    def test_gt(self):
        self._test_compare(gt_usecase)

    def test_identity(self):
        pyfunc = identity_usecase
        cfunc = jit(nopython=True)(pyfunc)
        self.assertPreciseEqual(cfunc(3), pyfunc(3))

    def test_bool_list(self):
        # Check lists of bools compile and run successfully
        pyfunc = bool_list_usecase
        cfunc = jit(nopython=True)(pyfunc)
        self.assertPreciseEqual(cfunc(), pyfunc())


class TestUnboxing(MemoryLeakMixin, TestCase):
    """
    Test unboxing of Python lists into native Numba lists.
    """

    @contextlib.contextmanager
    def assert_type_error(self, msg):
        with self.assertRaises(TypeError) as raises:
            yield
        if msg is not None:
            self.assertRegexpMatches(str(raises.exception), msg)

    def check_unary(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        def check(arg):
            expected = pyfunc(arg)
            got = cfunc(arg)
            self.assertPreciseEqual(got, expected)
        return check

    def test_numbers(self):
        check = self.check_unary(unbox_usecase)
        check([1, 2])
        check([1j, 2.5j])

    def test_tuples(self):
        check = self.check_unary(unbox_usecase2)
        check([(1, 2), (3, 4)])
        check([(1, 2j), (3, 4j)])
        check([(), (), ()])

    def test_list_inside_tuple(self):
        check = self.check_unary(unbox_usecase3)
        check((1, [2, 3, 4]))

    def test_list_of_tuples_inside_tuple(self):
        check = self.check_unary(unbox_usecase4)
        check((1, [(2,), (3,)]))

    def test_errors(self):
        # See #1545 and #1594: error checking should ensure the list is
        # homogeneous
        msg = "can't unbox heterogeneous list"
        pyfunc = noop
        cfunc = jit(nopython=True)(pyfunc)
        lst = [1, 2.5]
        with self.assert_type_error(msg):
            cfunc(lst)
        # The list hasn't been changed (bogus reflecting)
        self.assertEqual(lst, [1, 2.5])
        with self.assert_type_error(msg):
            cfunc([1, 2j])
        # Same when the list is nested in a tuple or namedtuple
        with self.assert_type_error(msg):
            cfunc((1, [1, 2j]))
        with self.assert_type_error(msg):
            cfunc(Point(1, [1, 2j]))
        # Issue #1638: tuples of different size.
        # Note the check is really on the tuple side.
        lst = [(1,), (2, 3)]
        with self.assertRaises(TypeError) as raises:
            cfunc(lst)
        msg = ("can't unbox heterogeneous list: "
                "UniTuple({0} x 1) != UniTuple({0} x 2)")
        self.assertEqual(str(raises.exception), msg.format(types.intp))



class TestListReflection(MemoryLeakMixin, TestCase):
    """
    Test reflection of native Numba lists on Python list objects.
    """

    def check_reflection(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        samples = [([1., 2., 3., 4.], [0.]),
                   ([1., 2., 3., 4.], [5., 6., 7., 8., 9.]),
                   ]
        for dest, src in samples:
            expected = list(dest)
            got = list(dest)
            pyres = pyfunc(expected, src)
            with self.assertRefCount(got, src):
                cres = cfunc(got, src)
                self.assertPreciseEqual(cres, pyres)
                self.assertPreciseEqual(expected, got)
                self.assertEqual(pyres[0] is expected, cres[0] is got)
                del pyres, cres

    def test_reflect_simple(self):
        self.check_reflection(reflect_simple)

    def test_reflect_conditional(self):
        self.check_reflection(reflect_conditional)

    def test_reflect_exception(self):
        """
        When the function exits with an exception, lists should still be
        reflected.
        """
        pyfunc = reflect_exception
        cfunc = jit(nopython=True)(pyfunc)
        l = [1, 2, 3]
        with self.assertRefCount(l):
            with self.assertRaises(ZeroDivisionError):
                cfunc(l)
            self.assertPreciseEqual(l, [1, 2, 3, 42])

    def test_reflect_same_list(self):
        """
        When the same list object is reflected twice, behaviour should
        be consistent.
        """
        pyfunc = reflect_dual
        cfunc = jit(nopython=True)(pyfunc)
        pylist = [1, 2, 3]
        clist = pylist[:]
        expected = pyfunc(pylist, pylist)
        got = cfunc(clist, clist)
        self.assertPreciseEqual(expected, got)
        self.assertPreciseEqual(pylist, clist)
        self.assertPreciseEqual(sys.getrefcount(pylist), sys.getrefcount(clist))

    def test_reflect_clean(self):
        """
        When the list wasn't mutated, no reflection should take place.
        """
        cfunc = jit(nopython=True)(noop)
        # Use a complex, as Python integers can be cached
        l = [12.5j]
        ids = [id(x) for x in l]
        cfunc(l)
        self.assertEqual([id(x) for x in l], ids)


class ManagedListTestCase(MemoryLeakMixin, TestCase):

    def assert_list_element_precise_equal(self, expect, got):
        self.assertEqual(len(expect), len(got))
        for a, b in zip(expect, got):
            self.assertPreciseEqual(a, b)


class TestListManagedElements(ManagedListTestCase):
    "Test list containing objects that need refct"

    def _check_element_equal(self, pyfunc):
        cfunc = jit(nopython=True)(pyfunc)
        con = [np.arange(3).astype(np.intp), np.arange(5).astype(np.intp)]
        expect = list(con)
        pyfunc(expect)
        got = list(con)
        cfunc(got)
        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_reflect_passthru(self):
        def pyfunc(con):
            pass
        self._check_element_equal(pyfunc)

    def test_reflect_appended(self):
        def pyfunc(con):
            con.append(np.arange(10).astype(np.intp))

        self._check_element_equal(pyfunc)

    def test_reflect_setitem(self):
        def pyfunc(con):
            con[1] = np.arange(10)

        self._check_element_equal(pyfunc)

    def test_reflect_popped(self):
        def pyfunc(con):
            con.pop()

        self._check_element_equal(pyfunc)

    def test_append(self):
        def pyfunc():
            con = []
            for i in range(300):
                con.append(np.arange(i, ).astype(np.intp))
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_append_noret(self):
        # This test make sure local dtor works
        def pyfunc():
            con = []
            for i in range(300):
                con.append(np.arange(i))
            c = 0.0
            for arr in con:
                c += arr.sum() / (1 + arr.size)
            return c

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assertEqual(expect, got)

    def test_reassign_refct(self):
        def pyfunc():
            con = []
            for i in range(5):
                con.append(np.arange(2))
            con[0] = np.arange(4)
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_get_slice(self):
        def pyfunc():
            con = []
            for i in range(5):
                con.append(np.arange(2))
            return con[2:4]

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_set_slice(self):
        def pyfunc():
            con = []
            for i in range(5):
                con.append(np.arange(2))
            con[1:3] = con[2:4]
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_pop(self):
        def pyfunc():
            con = []
            for i in range(20):
                con.append(np.arange(i + 1))
            while len(con) > 2:
                con.pop()
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_pop_loc(self):
        def pyfunc():
            con = []
            for i in range(1000):
                con.append(np.arange(i + 1))
            while len(con) > 2:
                con.pop(1)
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_del_range(self):
        def pyfunc():
            con = []
            for i in range(20):
                con.append(np.arange(i + 1))
            del con[3:10]
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )

    def test_list_of_list(self):
        def pyfunc():
            con = []
            for i in range(10):
                con.append([0] * i)
            return con

        cfunc = jit(nopython=True)(pyfunc)
        expect = pyfunc()
        got = cfunc()

        self.assertEqual(expect, got)



def expect_reflection_failure(fn):
    def wrapped(self, *args, **kwargs):
        self.disable_leak_check()
        with self.assertRaises(TypeError) as raises:
            fn(self, *args, **kwargs)
        expect_msg = 'cannot reflect element of reflected container'
        self.assertIn(expect_msg, str(raises.exception))

    return wrapped


class TestListOfList(ManagedListTestCase):

    def compile_and_test(self, pyfunc, *args):
        from copy import deepcopy
        expect_args = deepcopy(args)
        expect = pyfunc(*expect_args)

        njit_args = deepcopy(args)
        cfunc = jit(nopython=True)(pyfunc)
        got = cfunc(*njit_args)

        self.assert_list_element_precise_equal(
            expect=expect, got=got
            )
        # Check reflection
        self.assert_list_element_precise_equal(
            expect=expect_args, got=njit_args
            )

    def test_returning_list_of_list(self):
        def pyfunc():
            a = [[np.arange(i)] for i in range(4)]
            return a

        self.compile_and_test(pyfunc)

    @expect_reflection_failure
    def test_heterogeneous_list_error(self):
        def pyfunc(x):
            return x[1]

        cfunc = jit(nopython=True)(pyfunc)
        l2 = [[np.zeros(i) for i in range(5)],
              [np.ones(i)+1j for i in range(5)]]
        l3 = [[np.zeros(i) for i in range(5)], [(1,)]]
        l4 = [[1], [{1}]]
        l5 = [[1], [{'a': 1}]]

        # TODO: this triggers a reflection error.
        # Remove this line when nested reflection is supported
        cfunc(l2)

        # error_cases
        with self.assertRaises(TypeError) as raises:
            cfunc(l2)

        self.assertIn(
            ("reflected list(array(float64, 1d, C)) != "
             "reflected list(array(complex128, 1d, C))"),
            str(raises.exception)
            )

        with self.assertRaises(TypeError) as raises:
            cfunc(l3)

        self.assertIn(
            ("reflected list(array(float64, 1d, C)) != "
             "reflected list((int64 x 1))"),
            str(raises.exception)
            )

        with self.assertRaises(TypeError) as raises:
            cfunc(l4)
        self.assertIn(
            "reflected list(int64) != reflected list(reflected set(int64))",
            str(raises.exception)
            )

        with self.assertRaises(ValueError) as raises:
            cfunc(l5)
        self.assertIn(
            "Cannot type list element of <class 'dict'>",
            str(raises.exception)
            )

    @expect_reflection_failure
    def test_list_of_list_reflected(self):
        def pyfunc(l1, l2):
            l1.append(l2)
            l1[-1].append(123)

        cfunc = jit(nopython=True)(pyfunc)
        l1 = [[0, 1], [2, 3]]
        l2 = [4, 5]
        expect = list(l1), list(l2)
        got = list(l1), list(l2)
        pyfunc(*expect)
        cfunc(*got)
        self.assertEqual(expect, got)

    @expect_reflection_failure
    def test_heterogeneous_list(self):
        def pyfunc(x):
            return x[1]

        l1 = [[np.zeros(i) for i in range(5)], [np.ones(i) for i in range(5)]]

        cfunc = jit(nopython=True)(pyfunc)
        l1_got = cfunc(l1)
        self.assertPreciseEqual(pyfunc(l1), l1_got)

    @expect_reflection_failure
    def test_c01(self):
        def bar(x):
            return x.pop()

        r = [[np.zeros(0)], [np.zeros(10)*1j]]
        # TODO: this triggers a reflection error.
        # Remove this line when nested reflection is supported
        self.compile_and_test(bar, r)

        with self.assertRaises(TypeError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn(
            ("reflected list(array(float64, 1d, C)) != "
             "reflected list(array(complex128, 1d, C))"),
            str(raises.exception),
            )

    def test_c02(self):
        def bar(x):
            x.append(x)
            return x

        r = [[np.zeros(0)]]

        with self.assertRaises(errors.TypingError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn(
            "Invalid use of BoundFunction(list.append",
            str(raises.exception),
            )

    def test_c03(self):
        def bar(x):
            f = x
            f[0] = 1
            return f

        r = [[np.arange(3)]]

        with self.assertRaises(errors.TypingError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn(
            "invalid setitem with value of {} to element of {}".format(
                typeof(1),
                typeof(r[0]),
                ),
            str(raises.exception),
        )

    def test_c04(self):
        def bar(x):
            f = x
            f[0][0] = 10
            return f

        r = [[np.arange(3)]]
        with self.assertRaises(errors.TypingError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn(
            "invalid setitem with value of {} to element of {}".format(
                typeof(10),
                typeof(r[0][0]),
                ),
            str(raises.exception),
            )

    @expect_reflection_failure
    def test_c05(self):
        def bar(x):
            f = x
            f[0][0] = np.array([x for x in np.arange(10).astype(np.intp)])
            return f

        r = [[np.arange(3).astype(np.intp)]]
        self.compile_and_test(bar, r)

    def test_c06(self):
        def bar(x):
            f = x
            f[0][0] = np.array([x + 1j for x in np.arange(10)])
            return f

        r = [[np.arange(3)]]
        with self.assertRaises(errors.TypingError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn("invalid setitem with value", str(raises.exception))

    @expect_reflection_failure
    def test_c07(self):
        self.disable_leak_check()

        def bar(x):
            return x[-7]

        r = [[np.arange(3)]]
        cfunc = jit(nopython=True)(bar)
        with self.assertRaises(IndexError) as raises:
            cfunc(r)
        self.assertIn("getitem out of range", str(raises.exception))

    def test_c08(self):
        self.disable_leak_check()

        def bar(x):
            x[5] = 7
            return x

        r = [1, 2, 3]
        cfunc = jit(nopython=True)(bar)
        with self.assertRaises(IndexError) as raises:
            cfunc(r)
        self.assertIn("setitem out of range", str(raises.exception))

    def test_c09(self):
        def bar(x):
            x[-2] = 7j
            return x

        r = [1, 2, 3]
        with self.assertRaises(errors.TypingError) as raises:
            self.compile_and_test(bar, r)
        self.assertIn("invalid setitem with value", str(raises.exception))

    @expect_reflection_failure
    def test_c10(self):
        def bar(x):
            x[0], x[1] = x[1], x[0]
            return x

        r = [[1, 2, 3], [4, 5, 6]]
        self.compile_and_test(bar, r)

    @expect_reflection_failure
    def test_c11(self):
        def bar(x):
            x[:] = x[::-1]
            return x

        r = [[1, 2, 3], [4, 5, 6]]
        self.compile_and_test(bar, r)

    def test_c12(self):
        def bar(x):
            del x[-1]
            return x

        r = [x for x in range(10)]
        self.compile_and_test(bar, r)


class Item(object):
    def __init__(self, many, scalar):
        self.many = many
        self.scalar = scalar


class Container(object):
    def __init__(self, n):
        self.data = [[np.arange(i).astype(np.float64)] for i in range(n)]

    def more(self, n):
        for i in range(n):
            self.data.append([np.arange(i).astype(np.float64)])


class TestListAndJitClasses(ManagedListTestCase):
    def make_jitclass_element(self):
        spec = [
            ('many', types.float64[:]),
            ('scalar', types.float64),
        ]
        JCItem = jitclass(spec)(Item)
        return JCItem

    def make_jitclass_container(self):
        spec = {
            'data': types.List(dtype=types.List(types.float64[::1])),
        }
        JCContainer = jitclass(spec)(Container)
        return JCContainer

    def assert_list_element_with_tester(self, tester, expect, got):
        for x, y in zip(expect, got):
            tester(x, y)

    def test_jitclass_instance_elements(self):
        JCItem = self.make_jitclass_element()

        def pyfunc(xs):
            xs[1], xs[0] = xs[0], xs[1]
            return xs

        def eq(x, y):
            self.assertPreciseEqual(x.many, y.many)
            self.assertPreciseEqual(x.scalar, y.scalar)

        cfunc = jit(nopython=True)(pyfunc)

        arg = [JCItem(many=np.random.random(n + 1), scalar=n * 1.2)
               for n in range(5)]

        expect_arg = list(arg)
        got_arg = list(arg)

        expect_res = pyfunc(expect_arg)
        got_res = cfunc(got_arg)

        self.assert_list_element_with_tester(eq, expect_arg, got_arg)
        self.assert_list_element_with_tester(eq, expect_res, got_res)

    def test_jitclass_containing_list(self):
        JCContainer = self.make_jitclass_container()

        expect = Container(n=4)
        got = JCContainer(n=4)
        self.assert_list_element_precise_equal(got.data, expect.data)
        expect.more(3)
        got.more(3)
        self.assert_list_element_precise_equal(got.data, expect.data)


if __name__ == '__main__':
    unittest.main()



# Python 3 syntax only use cases, used in test_extending.py

# arg name is different, and there's no arg name to match before *args


def impl4(z, *args, kw=None):
    if z > 10:
        return 1
    else:
        return -1

# arg name is different but at a detectable location, with *args


def impl5(z, b, *args, kw=None):
    if z > 10:
        return 1
    else:
        return -1


def var_positional_impl(a, *star_args_token, kw=None, kw1=12):
    def impl(a, b, f, kw=None, kw1=12):
        if a > 10:
            return 1
        else:
            return -1
    return impl



import sys
import unittest
import warnings
import subprocess
from importlib import import_module
from contextlib import contextmanager
from types import ModuleType

from numba.core.errors import NumbaDeprecationWarning, _MovedModule
from numba.tests.support import TestCase
from numba.core.utils import PYVERSION

IS_PY36 = PYVERSION[:2] == (3, 6)
NOT_PY36 = not IS_PY36

# All pre0.49 top-level module re-export that will be removed.
_pre_49_top_level_modules = [
    'analysis',
    'annotations',
    'appdirs',
    'array_analysis',
    'bytecode',
    'byteflow',
    'caching',
    'callwrapper',
    'cgutils',
    'charseq',
    'compiler',
    'compiler_lock',
    'compiler_machinery',
    'consts',
    'controlflow',
    # 'ctypes_support', # removed
    'dataflow',
    'datamodel',
    'debuginfo',
    'decorators',
    'dictobject',
    'dispatcher',
    'entrypoints',
    'funcdesc',
    'generators',
    'inline_closurecall',
    'interpreter',
    # 'io_support', # removed
    'ir',
    'ir_utils',
    'itanium_mangler',
    'listobject',
    'lowering',
    'npdatetime',
    # 'npyufunc', # removed
    'numpy_support',
    'object_mode_passes',
    'parfor',
    'postproc',
    'pylowering',
    'pythonapi',
    'rewrites',
    'runtime',
    'serialize',
    'sigutils',
    # 'six', # removed
    'special',
    'stencilparfor',
    # 'targets', # removed
    'tracing',
    'transforms',
    'typeconv',
    'typed_passes',
    'typedobjectutils',
    'typeinfer',
    'typing',
    'unicode',
    'unicode_support',
    'unsafe',
    'untyped_passes',
    'utils',
    'withcontexts',
]


class TestAPIMoves_Q1_2020(TestCase):
    """ Checks API moves made in Q1 2020, this roughly corresponds to 0.48->0.49
    """

    def check_warning(self, from_mod, to_mod):
        """
        Returns a context manager to check that a NumbaDeprecationWarning is
        raised in the context managed block and that the warning contains
        information that from_mod has moved to to_mod
        """
        # make sure the `to_mod` location actually exists (None indicates no
        # replacement"
        if to_mod is not None:
            import_module(to_mod)

        @contextmanager
        def checker(fn=None, skip=False):
            """
            If fn is not None then a check will be made to ensure the module
            level `__getattr__`
            """
            with warnings.catch_warnings(record=True) as w:
                warnings.simplefilter("always", NumbaDeprecationWarning)
                yield
            if skip:
                return
            self.assertTrue(len(w) > 0)
            _require = "requested from a module that has moved location"
            for x in w:
                if to_mod is not None:
                    c1 = _require in str(x.message)
                    c2 = from_mod in str(x.message)
                    c3 = to_mod in str(x.message)
                else:  # check for help link
                    c1 = True
                    c2 = from_mod in str(x.message)
                    c3 = "gitter.im" in str(x.message)
                if fn is not None:
                    c4 = "Import of '{}' requested".format(fn) in str(
                        x.message
                    )
                else:
                    c4 = True
                if c1 and c2 and c3 and c4:
                    break
            else:
                raise ValueError("No valid warning message found")

        return checker

    def test_nonexistant_moved_raises_not_warns(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDeprecationWarning)
            with self.assertRaises(ImportError) as raises:
                from numba import nonexistant_thing  # noqa: F401
            self.assertIn("cannot import name", str(raises.exception))
            self.assertIn("nonexistant_thing", str(raises.exception))
        self.assertEqual(len(w), 0)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDeprecationWarning)
            with self.assertRaises(ImportError) as raises:
                import numba.nonexistant_thing  # noqa: F401
            self.assertIn("No module named", str(raises.exception))
            self.assertIn("nonexistant_thing", str(raises.exception))
        self.assertEqual(len(w), 0)

    def test_numba_utils(self):
        checker = self.check_warning("numba.utils", "numba.core.utils")
        with checker(skip=IS_PY36):
            import numba.utils

        for fn in ("pysignature", "OPERATORS_TO_BUILTINS"):
            with checker(fn):
                getattr(numba.utils, fn)

    def test_numba_untyped_passes(self):
        checker = self.check_warning(
            "numba.untyped_passes", "numba.core.untyped_passes"
        )
        with checker(skip=IS_PY36):
            import numba.untyped_passes

        fn = "InlineClosureLikes"
        with checker(fn):
            getattr(numba.untyped_passes, fn)

    def test_numba_config(self):
        checker = self.check_warning(
            "numba.config", "numba.core.config"
        )
        with checker(skip=IS_PY36):
            import numba.config

        fn = "DISABLE_JIT"
        with checker(fn):
            getattr(numba.config, fn)

    def test_numba_unsafe(self):
        # this is a bit dubious, unsafe was largely split up and moved.
        # numba.unsafe.refcount is in use in external packages and is now in
        # core
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDeprecationWarning)
            fn = "refcount"
            import numba.unsafe

            getattr(numba.unsafe, fn)

        if NOT_PY36:
            for x in w:
                if "No direct replacement for 'numba.unsafe'" in str(x.message):
                    break
            else:
                self.fail("Could not find expected warning message")

    def test_numba_unsafe_ndarray(self):
        checker = self.check_warning(
            "numba.unsafe.ndarray", "numba.np.unsafe.ndarray"
        )
        with checker():
            import numba.unsafe.ndarray

        fn = "empty_inferred"
        with checker(fn):
            getattr(numba.unsafe.ndarray, fn)

    def test_numba_unicode(self):
        checker = self.check_warning("numba.unicode", "numba.cpython.unicode")
        with checker(skip=IS_PY36):
            import numba.unicode

        for fn in (
            "unbox_unicode_str",
            "make_string_from_constant",
            "_slice_span",
            "_normalize_slice",
            "_empty_string",
            "PY_UNICODE_1BYTE_KIND",
            "PY_UNICODE_2BYTE_KIND",
            "PY_UNICODE_4BYTE_KIND",
            "PY_UNICODE_WCHAR_KIND",
        ):
            with checker(fn):
                getattr(numba.unicode, fn)

    def test_aaaaa_numba_typing(self):
        # silly 'aaaaa' name to game test ordering, warnings only get triggered
        # once and the `TestAPIMoves_Q1_2020` hits numba.typing.* so this needs
        # to run first
        checker = self.check_warning("numba.typing", "numba.core.typing")

        with checker(skip=IS_PY36):
            import numba.typing

        for fn in (
            "signature",
            "make_concrete_template",
            "Signature",
            "fold_arguments",
        ):
            with checker(fn):
                getattr(numba.typing, fn)

    def test_numba_typing_typeof(self):
        checker = self.check_warning(
            "numba.typing.typeof", "numba.core.typing.typeof"
        )
        with checker():
            import numba.typing.typeof

        for fn in ("typeof_impl", "_typeof_ndarray"):
            with checker(fn):
                getattr(numba.typing.typeof, fn)

    def test_numba_typing_templates(self):
        checker = self.check_warning(
            "numba.typing.templates", "numba.core.typing.templates"
        )
        with checker():
            import numba.typing.templates

        for fn in (
            "signature",
            "infer_global",
            "infer_getattr",
            "infer",
            "bound_function",
            "AttributeTemplate",
            "AbstractTemplate",
        ):
            with checker(fn):
                getattr(numba.typing.templates, fn)

    def test_numba_typing_npydecl(self):
        checker = self.check_warning(
            "numba.typing.npydecl", "numba.core.typing.npydecl"
        )
        with checker():
            import numba.typing.npydecl

        for fn in (
            "supported_ufuncs",
            "NumpyRulesInplaceArrayOperator",
            "NumpyRulesArrayOperator",
            "NdConcatenate",
        ):
            with checker(fn):
                getattr(numba.typing.npydecl, fn)

    def test_numba_typing_ctypes_utils(self):
        checker = self.check_warning(
            "numba.typing.ctypes_utils", "numba.core.typing.ctypes_utils"
        )
        with checker():
            import numba.typing.ctypes_utils

        fn = "make_function_type"
        with checker(fn):
            getattr(numba.typing.ctypes_utils, fn)

    def test_numba_typing_collections(self):
        checker = self.check_warning(
            "numba.typing.collections", "numba.core.typing.collections"
        )
        with checker():
            import numba.typing.collections

        fn = "GetItemSequence"
        with checker(fn):
            getattr(numba.typing.collections, fn)

    def test_numba_typing_builtins(self):
        checker = self.check_warning(
            "numba.typing.builtins", "numba.core.typing.builtins"
        )
        with checker():
            import numba.typing.builtins

        for fn in (
            "MinMaxBase",
            "IndexValueType",
        ):
            with checker(fn):
                getattr(numba.typing.builtins, fn)

    def test_numba_typing_arraydecl(self):
        checker = self.check_warning(
            "numba.typing.arraydecl", "numba.core.typing.arraydecl"
        )
        with checker():
            import numba.typing.arraydecl

        for fn in (
            "get_array_index_type",
            "ArrayAttribute",
        ):
            with checker(fn):
                getattr(numba.typing.arraydecl, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.typing.arraydecl import ArrayAttribute

            ArrayAttribute.resolve_var
            ArrayAttribute.resolve_sum
            ArrayAttribute.resolve_prod

    def test_numba_typeinfer(self):
        checker = self.check_warning("numba.typeinfer", "numba.core.typeinfer")
        with checker(skip=IS_PY36):
            import numba.typeinfer

        fn = "IntrinsicCallConstraint"
        with checker(fn):
            getattr(numba.typeinfer, fn)

    def test_numba_typedobjectutils(self):
        checker = self.check_warning(
            "numba.typedobjectutils", "numba.typed.typedobjectutils"
        )
        with checker(skip=IS_PY36):
            import numba.typedobjectutils

        fn = "_cast"
        with checker(fn):
            getattr(numba.typedobjectutils, fn)

    def test_numba_typed_passes(self):
        checker = self.check_warning(
            "numba.typed_passes", "numba.core.typed_passes"
        )
        with checker(skip=IS_PY36):
            import numba.typed_passes

        for fn in ("type_inference_stage", "AnnotateTypes"):
            with checker(fn):
                getattr(numba.typed_passes, fn)

    def test_numba_typed(self):
        # no refactoring was done to `numba.typed`.
        import numba.typed  # noqa: F401
        from numba.typed import List

        List.empty_list
        from numba.typed import Dict

        Dict.empty

    def test_numba_typeconv(self):
        checker = self.check_warning("numba.typeconv", "numba.core.typeconv")
        with checker(skip=IS_PY36):
            import numba.typeconv

        fn = "Conversion"
        with checker(fn):
            getattr(numba.typeconv, fn)

    def test_numba_types(self):
        checker = self.check_warning("numba.types", "numba.core.types")
        with checker():
            import numba.types

        for fn in ("int64", "float64"):
            with checker(fn):
                getattr(numba.types, fn)

    def test_aaaaa_numba_targets(self):
        # silly 'aaaaa' name to game test ordering, warnings only get triggered
        # once and the `TestAPIMoves_Q1_2020` hits numba.targets.* so this needs
        # to run first
        checker = self.check_warning("numba.targets", None)
        with checker():
            import numba.targets  # noqa: F401

    def test_numba_targets_boxing(self):
        checker = self.check_warning(
            "numba.targets.boxing", "numba.core.boxing"
        )
        with checker():
            import numba.targets
            numba.targets.boxing

        for fn in ("box_array", "_NumbaTypeHelper"):
            with checker(fn):
                getattr(numba.targets.boxing, fn)

    def test_numba_targets_callconv(self):
        checker = self.check_warning(
            "numba.targets.callconv", "numba.core.callconv"
        )
        with checker():
            import numba.targets
            numba.targets.callconv

        fn = "RETCODE_USEREXC"
        with checker(fn):
            getattr(numba.targets.callconv, fn)

    def test_numba_targets_hashing(self):
        checker = self.check_warning(
            "numba.targets.hashing", "numba.cpython.hashing"
        )
        with checker():
            import numba.targets
            numba.targets.hashing

        # some codebases claim to use these, but they are not present in 0.48:
        # '_Py_HashSecret_siphash_k0', '_Py_HashSecret_siphash_k1',
        # '_Py_HashSecret_djbx33a_suffix'
        fn = "_Py_hash_t"
        with checker(fn):
            getattr(numba.targets.hashing, fn)

    def test_numba_targets_ufunc_db(self):
        checker = self.check_warning(
            "numba.targets.ufunc_db", "numba.np.ufunc_db"
        )
        with checker():
            import numba.targets
            numba.targets.ufunc_db

        fn = "get_ufuncs"
        with checker(fn):
            getattr(numba.targets.ufunc_db, fn)

    def test_numba_targets_slicing(self):
        checker = self.check_warning(
            "numba.targets.slicing", "numba.cpython.slicing"
        )
        with checker():
            import numba.targets
            numba.targets.slicing

        for fn in ("guard_invalid_slice", "get_slice_length", "fix_slice"):
            with checker(fn):
                getattr(numba.targets.slicing, fn)

    def test_numba_targets_setobj(self):
        checker = self.check_warning(
            "numba.targets.setobj", "numba.cpython.setobj"
        )
        with checker():
            import numba.targets
            numba.targets.setobj

        fn = "set_empty_constructor"
        with checker(fn):
            getattr(numba.targets.setobj, fn)

    def test_numba_targets_registry(self):
        checker = self.check_warning(
            "numba.targets.registry", "numba.core.registry"
        )
        with checker():
            import numba.targets
            numba.targets.registry

        for fn in ("dispatcher_registry", "cpu_target"):
            with checker(fn):
                getattr(numba.targets.registry, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.targets.registry import cpu_target

            typing_context = cpu_target.typing_context
            typing_context.resolve_value_type

    def test_numba_targets_options(self):
        checker = self.check_warning(
            "numba.targets.options", "numba.core.options"
        )
        with checker():
            import numba.targets
            numba.targets.options

        fn = "TargetOptions"
        with checker(fn):
            getattr(numba.targets.options, fn)

    def test_numba_targets_listobj(self):
        checker = self.check_warning(
            "numba.targets.listobj", "numba.cpython.listobj"
        )
        with checker():
            import numba.targets
            numba.targets.listobj

        fn = "ListInstance"
        with checker(fn):
            getattr(numba.targets.listobj, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.targets.listobj import ListInstance

            ListInstance.allocate

    def test_numba_targets_imputils(self):
        checker = self.check_warning(
            "numba.targets.imputils", "numba.core.imputils"
        )
        with checker():
            import numba.targets
            numba.targets.imputils

        for fn in (
            "lower_cast",
            "iternext_impl",
            "impl_ret_new_ref",
            "RefType",
        ):
            with checker(fn):
                getattr(numba.targets.imputils, fn)

    def test_numba_targets_cpu(self):
        checker = self.check_warning("numba.targets.cpu", "numba.core.cpu")
        with checker():
            import numba.targets
            numba.targets.cpu

        for fn in ("ParallelOptions", "CPUTargetOptions", "CPUContext"):
            with checker(fn):
                getattr(numba.targets.cpu, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.targets.cpu import CPUTargetOptions

            CPUTargetOptions.OPTIONS

    def test_numba_targets_builtins(self):
        checker = self.check_warning(
            "numba.targets.builtins", "numba.cpython.builtins"
        )
        with checker():
            import numba.targets
            numba.targets.builtins

        for fn in ("get_type_min_value", "get_type_max_value"):
            with checker(fn):
                getattr(numba.targets.builtins, fn)

    def test_numba_targets_arrayobj(self):
        checker = self.check_warning(
            "numba.targets.arrayobj", "numba.np.arrayobj"
        )
        with checker():
            import numba.targets
            numba.targets.arrayobj

        for fn in (
            "store_item",
            "setitem_array",
            "populate_array",
            "numpy_empty_nd",
            "make_array",
            "getiter_array",
            "getitem_arraynd_intp",
            "getitem_array_tuple",
            "fancy_getitem_array",
            "array_reshape",
            "array_ravel",
            "array_len",
            "array_flatten",
        ):
            with checker(fn):
                getattr(numba.targets.arrayobj, fn)

    def test_numba_targets_arraymath(self):
        checker = self.check_warning(
            "numba.targets.arraymath", "numba.np.arraymath"
        )
        with checker():
            import numba.targets
            numba.targets.arraymath

        fn = "get_isnan"
        with checker(fn):
            getattr(numba.targets.arraymath, fn)

    def test_numba_targets_nonexistant(self):
        # not testing for this warning
        with warnings.catch_warnings(record=True):
            warnings.simplefilter("always", NumbaDeprecationWarning)
            import numba.targets

        with self.assertRaises(AttributeError) as raises:
            numba.targets.does_not_exist
        exc = raises.exception
        self.assertIn("no attribute or submodule", str(exc))

        # chain shows both the moved attribute lookup and import failing
        self.assertIsInstance(exc.__context__, ModuleNotFoundError)
        self.assertIsInstance(exc.__context__.__context__, AttributeError)

    def test_numba_stencilparfor(self):
        checker = self.check_warning(
            "numba.stencilparfor", "numba.stencils.stencilparfor"
        )
        with checker(skip=IS_PY36):
            import numba.stencilparfor

        fn = "_compute_last_ind"
        with checker(fn):
            getattr(numba.stencilparfor, fn)

    def test_numba_stencil(self):
        checker = self.check_warning("numba.stencil", "numba.stencils.stencil")
        with checker():
            import numba.stencil

        fn = "StencilFunc"
        with checker(fn):
            getattr(numba.stencil, fn)

    def test_numba_runtime_nrt(self):
        checker = self.check_warning(
            "numba.runtime.nrt", "numba.core.runtime.nrt"
        )
        with checker():
            import numba.runtime.nrt

        fn = "rtsys"
        with checker(fn):
            getattr(numba.runtime.nrt, fn)

    def test_numba_runtime(self):
        checker = self.check_warning("numba.runtime", "numba.core.runtime")
        with checker(skip=IS_PY36):
            import numba.runtime

        fn = "nrt"
        with checker(fn):
            getattr(numba.runtime, fn)

    def test_numba_rewrites(self):
        checker = self.check_warning("numba.rewrites", "numba.core.rewrites")
        with checker(skip=IS_PY36):
            import numba.rewrites

        fn = "rewrite_registry"
        with checker(skip=IS_PY36):
            getattr(numba.rewrites, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.rewrites import rewrite_registry

            rewrite_registry.apply

    def test_numba_pythonapi(self):
        checker = self.check_warning("numba.pythonapi", "numba.core.pythonapi")
        with checker(skip=IS_PY36):
            import numba.pythonapi

        for fn in (
            "_UnboxContext",
            "_BoxContext",
        ):
            with checker(fn):
                getattr(numba.pythonapi, fn)

    def test_numba_parfor(self):
        checker = self.check_warning("numba.parfor", "numba.parfors.parfor")
        with checker(skip=IS_PY36):
            import numba.parfor

        for fn in (
            "replace_functions_map",
            "min_checker",
            "maximize_fusion",
            "max_checker",
            "lower_parfor_sequential",
            "internal_prange",
            "init_prange",
            "argmin_checker",
            "argmax_checker",
            "PreParforPass",
            "ParforPass",
            "Parfor",
        ):
            with checker(fn):
                getattr(numba.parfor, fn)

    def test_numba_numpy_support(self):
        checker = self.check_warning(
            "numba.numpy_support", "numba.np.numpy_support"
        )
        if NOT_PY36:
            with checker():
                import numba.numpy_support
        import numba
        for fn in ("map_layout", "from_dtype", "as_dtype"):
            with checker(fn):
                getattr(numba.numpy_support, fn)

    def test_numba_targets_npdatetime(self):
        checker = self.check_warning(
            "numba.targets.npdatetime", "numba.np.npdatetime"
        )
        with checker():
            import numba.targets
            numba.targets.npdatetime

        for fn in ("convert_datetime_for_arith",):
            with checker(fn):
                getattr(numba.targets.npdatetime, fn)

    def test_numba_npdatetime(self):
        #  the <= 0.48  numba.npdatetime module provides the datetime helper
        # functions, this got moved in location and name
        checker = self.check_warning(
            "numba.npdatetime", "numba.np.npdatetime_helpers"
        )
        with checker(skip=IS_PY36):
            import numba.npdatetime

        for fn in ("DATETIME_UNITS",):
            with checker(fn):
                getattr(numba.npdatetime, fn)

    def test_numba_lowering(self):
        checker = self.check_warning("numba.lowering", "numba.core.lowering")
        with checker(skip=IS_PY36):
            import numba.lowering  # noqa: F401

    def test_numba_ir_utils(self):
        checker = self.check_warning("numba.ir_utils", "numba.core.ir_utils")
        with checker(skip=IS_PY36):
            import numba.ir_utils

        for fn in (
            "simplify_CFG",
            "replace_vars_stmt",
            "replace_arg_nodes",
            "remove_dead",
            "remove_call_handlers",
            "next_label",
            "mk_unique_var",
            "get_ir_of_code",
            "get_definition",
            "find_const",
            "dprint_func_ir",
            "compute_cfg_from_blocks",
            "compile_to_numba_ir",
            "build_definitions",
            "alias_func_extensions",
            "_max_label",
            "_add_alias",
            "GuardException",
        ):
            with checker(fn):
                getattr(numba.ir_utils, fn)

    def test_numba_ir(self):
        checker = self.check_warning("numba.ir", "numba.core.ir")
        with checker(skip=IS_PY36):
            import numba.ir

        for fn in ("Assign", "Const", "Expr", "Var"):
            with checker(fn):
                getattr(numba.ir, fn)

    def test_numba_inline_closurecall(self):
        checker = self.check_warning(
            "numba.inline_closurecall", "numba.core.inline_closurecall"
        )
        with checker(skip=IS_PY36):
            import numba.inline_closurecall

        for fn in (
            "inline_closure_call",
            "_replace_returns",
            "_replace_freevars",
            "_add_definitions",
        ):
            with checker(fn):
                getattr(numba.inline_closurecall, fn)

    def test_numba_errors(self):
        checker = self.check_warning("numba.errors", "numba.core.errors")
        with checker(skip=IS_PY36):
            import numba.errors

        for fn in (
            "WarningsFixer",
            "TypingError",
            "NumbaWarning",
            "ForceLiteralArg",
        ):
            with checker(fn):
                getattr(numba.errors, fn)

    def test_numba_dispatcher(self):
        checker = self.check_warning(
            "numba.dispatcher", "numba.core.dispatcher"
        )
        with checker(skip=IS_PY36):
            import numba.dispatcher

        for fn in (
            "ObjModeLiftedWith",
            "Dispatcher",
        ):
            with checker(fn):
                getattr(numba.dispatcher, fn)

    def test_numba_dictobject(self):
        checker = self.check_warning(
            "numba.dictobject", "numba.typed.dictobject"
        )
        with checker(skip=IS_PY36):
            import numba.dictobject

        fn = "DictModel"
        with checker(fn):
            getattr(numba.dictobject, fn)

    def test_numba_datamodel(self):
        checker = self.check_warning("numba.datamodel", "numba.core.datamodel")
        with checker(skip=IS_PY36):
            import numba.datamodel

        for fn in (
            "registry",
            "register_default",
        ):
            with checker(fn):
                getattr(numba.datamodel, fn)

        checker = self.check_warning(
            "numba.datamodel.models", "numba.core.datamodel.models"
        )
        with checker():
            import numba.datamodel.models

        for fn in ("StructModel", "PointerModel", "BooleanModel"):
            with checker(fn):
                getattr(numba.datamodel.models, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.datamodel import registry

            registry.register_default

    def test_numba_compiler_machinery(self):
        checker = self.check_warning(
            "numba.compiler_machinery", "numba.core.compiler_machinery"
        )
        with checker(skip=IS_PY36):
            import numba.compiler_machinery  # noqa: F401

    def test_numba_compiler(self):
        checker = self.check_warning("numba.compiler", "numba.core.compiler")
        with checker(skip=IS_PY36):
            import numba.compiler

        for fn in (
            "run_frontend",
            "StateDict",
            "Flags",
            "DefaultPassBuilder",
            "CompilerBase",
        ):
            with checker(fn):
                getattr(numba.compiler, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.compiler import DefaultPassBuilder, Flags

            DefaultPassBuilder.define_nopython_pipeline
            Flags.OPTIONS

    def test_numba_cgutils(self):
        checker = self.check_warning("numba.cgutils", "numba.core.cgutils")
        with checker(skip=IS_PY36):
            import numba.cgutils
        for fn in (
            "unpack_tuple",
            "true_bit",
            "is_not_null",
            "increment_index",
            "get_null_value",
            "false_bit",
            "create_struct_proxy",
            "alloca_once_value",
            "alloca_once",
        ):
            with checker(fn):
                getattr(numba.cgutils, fn)

    def test_numba_array_analysis(self):
        checker = self.check_warning(
            "numba.array_analysis", "numba.parfors.array_analysis"
        )
        with checker(skip=IS_PY36):
            import numba.array_analysis

        for fn in ("ArrayAnalysis", "array_analysis_extensions"):
            with checker(fn):
                getattr(numba.array_analysis, fn)

        # want to check this works, not that it warns
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", NumbaDeprecationWarning)
            from numba.array_analysis import ArrayAnalysis

            args = (None,) * 4
            ArrayAnalysis(*args)._analyze_op_static_getitem
            ArrayAnalysis(*args)._analyze_broadcast

    def test_numba_analysis(self):
        checker = self.check_warning("numba.analysis", "numba.core.analysis")
        if NOT_PY36:
            with checker():
                import numba.analysis
        import numba
        for fn in (
            "ir_extension_usedefs",
            "compute_use_defs",
            "compute_cfg_from_blocks",
            "_use_defs_result",
        ):
            with checker(fn):
                getattr(numba.analysis, fn)

    def test_numba_decorators(self):
        checker = self.check_warning(
            "numba.decorators", "numba.core.decorators"
        )
        with checker(skip=IS_PY36):
            import numba.decorators
        for fn in (
            "njit",
            "jit",
            "generated_jit",
        ):
            with checker(fn):
                getattr(numba.decorators, fn)

    def test_numba_jitclass(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDeprecationWarning)
            from numba import jitclass # SHOULD NOT WARN
        for x in w:
            if "numba.jitclass" in str(x.message):
                raise ValueError("Unexpected warning encountered")

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always", NumbaDeprecationWarning)
            @jitclass(spec=[])
            class foo:
                def __init__(self):
                    pass
            foo()
        for x in w:
            if "has moved to 'numba.experimental.jitclass'" in str(x.message):
                break
        else:
            raise ValueError("Expected warning not found")

    def run_in_fresh_python(self, code):
        return subprocess.check_output(
            [sys.executable, '-c', code],
            stderr=subprocess.STDOUT,
        ).decode()

    def test_top_level_export(self):
        def compare_mod(left, right):
            if isinstance(left, _MovedModule):
                self.assertIsInstance(right, ModuleType)
            elif isinstance(right, _MovedModule):
                self.assertIsInstance(left, ModuleType)
            else:
                self.assertIs(left, right)

        for mod in _pre_49_top_level_modules:
            if mod in {'unsafe'}:
                # SKIP.
                # different behavior and tested elsewhere.
                continue
            oldmodname = f'numba.{mod}'
            with self.subTest(oldmodname):
                oldmod = import_module(oldmodname)
                # Make sure the target module is available
                newmodname = oldmod._MovedModule__new_module
                newmod = import_module(newmodname)
                for k in dir(newmod):
                    if k.startswith('_'):
                        continue
                    obj = getattr(oldmod, k)
                    compare_mod(getattr(newmod, k), obj)

    def test_top_level_export_as_attr(self):
        for mod in _pre_49_top_level_modules:
            mod = f'numba.{mod}'
            with self.subTest(mod):
                out = self.run_in_fresh_python(f'import numba; {mod}')
                if NOT_PY36:
                    self.assertIn("NumbaDeprecationWarning", out)
                    self.assertIn(f"Import requested from: '{mod}'", out)

    def test_top_level_export_as_import(self):
        for mod in _pre_49_top_level_modules:
            if mod in {'unsafe'}:
                # SKIP.
                # different behavior and tested elsewhere.
                continue
            mod = f'numba.{mod}'
            with self.subTest(mod):
                out = self.run_in_fresh_python(f'import {mod}')
                if NOT_PY36:
                    self.assertIn("NumbaDeprecationWarning", out)
                    self.assertIn(f"Import requested from: '{mod}'", out)

    def test_top_level_export_attr_matches_import(self):
        for mod in _pre_49_top_level_modules:
            mod = f'numba.{mod}'
            with self.subTest(mod):
                code = [
                    'import numba',
                    f'import {mod} as ref',
                    f'assert {mod} is ref',
                ]
                self.run_in_fresh_python('; '.join(code))

    def test_issue_5528(self):
        # Check specific issues from the ticket.
        code = """
import numba
numba.numpy_support.from_dtype
"""
        self.run_in_fresh_python(code)

        code = """
import numba.numpy_support
numba.numpy_support.from_dtype
"""
        self.run_in_fresh_python(code)

        code = """
from numba.numpy_support import from_dtype
"""
        self.run_in_fresh_python(code)

        code = """
from numba.targets import quicksort
"""
        self.run_in_fresh_python(code)

        code = """
import numba
numba.special
"""
        self.run_in_fresh_python(code)

        code = """
from numba.special import literally
"""
        self.run_in_fresh_python(code)

        code = """
from numba.special import prange
"""
        self.run_in_fresh_python(code)

        code = """
from numba import postproc
"""
        self.run_in_fresh_python(code)

        code = """
import numba
numba.analysis.ir_extension_usedefs

"""
        self.run_in_fresh_python(code)

        code = """
from numba import consts
"""
        self.run_in_fresh_python(code)

        code = """
import numba
numba.extending.lower_builtin

"""
        self.run_in_fresh_python(code)

        code = """
from numba.config import IS_32BITS
"""
        self.run_in_fresh_python(code)

        code = """
import numba
numba.datamodel.register_default
"""
        self.run_in_fresh_python(code)


if __name__ == "__main__":
    unittest.main()



"""
Implementation of operations on numpy timedelta64.
"""

import numpy as np
import operator

from llvmlite.llvmpy.core import Type, Constant
import llvmlite.llvmpy.core as lc

from numba.core import types, cgutils
from numba.core.imputils import (lower_builtin, lower_constant,
                                    impl_ret_untracked)
from numba.np import npdatetime_helpers, numpy_support, npyfuncs

# datetime64 and timedelta64 use the same internal representation
DATETIME64 = TIMEDELTA64 = Type.int(64)
NAT = Constant.int(TIMEDELTA64, npdatetime_helpers.NAT)

TIMEDELTA_BINOP_SIG = (types.NPTimedelta,) * 2


def scale_by_constant(builder, val, factor):
    """
    Multiply *val* by the constant *factor*.
    """
    return builder.mul(val, Constant.int(TIMEDELTA64, factor))


def unscale_by_constant(builder, val, factor):
    """
    Divide *val* by the constant *factor*.
    """
    return builder.sdiv(val, Constant.int(TIMEDELTA64, factor))


def add_constant(builder, val, const):
    """
    Add constant *const* to *val*.
    """
    return builder.add(val, Constant.int(TIMEDELTA64, const))


def scale_timedelta(context, builder, val, srcty, destty):
    """
    Scale the timedelta64 *val* from *srcty* to *destty*
    (both numba.types.NPTimedelta instances)
    """
    factor = npdatetime_helpers.get_timedelta_conversion_factor(
        srcty.unit, destty.unit)
    if factor is None:
        # This can happen when using explicit output in a ufunc.
        raise NotImplementedError("cannot convert timedelta64 from %r to %r"
                                  % (srcty.unit, destty.unit))
    return scale_by_constant(builder, val, factor)


def normalize_timedeltas(context, builder, left, right, leftty, rightty):
    """
    Scale either *left* or *right* to the other's unit, in order to have
    homogeneous units.
    """
    factor = npdatetime_helpers.get_timedelta_conversion_factor(
        leftty.unit, rightty.unit)
    if factor is not None:
        return scale_by_constant(builder, left, factor), right
    factor = npdatetime_helpers.get_timedelta_conversion_factor(
        rightty.unit, leftty.unit)
    if factor is not None:
        return left, scale_by_constant(builder, right, factor)
    # Typing should not let this happen, except on == and != operators
    raise RuntimeError("cannot normalize %r and %r" % (leftty, rightty))


def alloc_timedelta_result(builder, name='ret'):
    """
    Allocate a NaT-initialized datetime64 (or timedelta64) result slot.
    """
    ret = cgutils.alloca_once(builder, TIMEDELTA64, name=name)
    builder.store(NAT, ret)
    return ret


def alloc_boolean_result(builder, name='ret'):
    """
    Allocate an uninitialized boolean result slot.
    """
    ret = cgutils.alloca_once(builder, Type.int(1), name=name)
    return ret


def is_not_nat(builder, val):
    """
    Return a predicate which is true if *val* is not NaT.
    """
    return builder.icmp(lc.ICMP_NE, val, NAT)


def are_not_nat(builder, vals):
    """
    Return a predicate which is true if all of *vals* are not NaT.
    """
    assert len(vals) >= 1
    pred = is_not_nat(builder, vals[0])
    for val in vals[1:]:
        pred = builder.and_(pred, is_not_nat(builder, val))
    return pred


def make_constant_array(vals):
    consts = [Constant.int(TIMEDELTA64, v) for v in vals]
    return Constant.array(TIMEDELTA64, consts)


normal_year_months = make_constant_array([31, 28, 31, 30, 31, 30,
                                          31, 31, 30, 31, 30, 31])
leap_year_months = make_constant_array([31, 29, 31, 30, 31, 30,
                                        31, 31, 30, 31, 30, 31])
normal_year_months_acc = make_constant_array(
    [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334])
leap_year_months_acc = make_constant_array(
    [0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335])


@lower_constant(types.NPDatetime)
@lower_constant(types.NPTimedelta)
def datetime_constant(context, builder, ty, pyval):
    return DATETIME64(pyval.astype(np.int64))


# Arithmetic operators on timedelta64

@lower_builtin(operator.pos, types.NPTimedelta)
def timedelta_pos_impl(context, builder, sig, args):
    res = args[0]
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.neg, types.NPTimedelta)
def timedelta_neg_impl(context, builder, sig, args):
    res = builder.neg(args[0])
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(abs, types.NPTimedelta)
def timedelta_abs_impl(context, builder, sig, args):
    val, = args
    ret = alloc_timedelta_result(builder)
    with builder.if_else(cgutils.is_scalar_neg(builder, val)) as (then, otherwise):
        with then:
            builder.store(builder.neg(val), ret)
        with otherwise:
            builder.store(val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


def timedelta_sign_impl(context, builder, sig, args):
    """
    np.sign(timedelta64)
    """
    val, = args
    ret = alloc_timedelta_result(builder)
    zero = Constant.int(TIMEDELTA64, 0)
    with builder.if_else(builder.icmp(lc.ICMP_SGT, val, zero)
                         ) as (gt_zero, le_zero):
        with gt_zero:
            builder.store(Constant.int(TIMEDELTA64, 1), ret)
        with le_zero:
            with builder.if_else(builder.icmp(lc.ICMP_EQ, val, zero)
                                 ) as (eq_zero, lt_zero):
                with eq_zero:
                    builder.store(Constant.int(TIMEDELTA64, 0), ret)
                with lt_zero:
                    builder.store(Constant.int(TIMEDELTA64, -1), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.add, *TIMEDELTA_BINOP_SIG)
@lower_builtin(operator.iadd, *TIMEDELTA_BINOP_SIG)
def timedelta_add_impl(context, builder, sig, args):
    [va, vb] = args
    [ta, tb] = sig.args
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = scale_timedelta(context, builder, va, ta, sig.return_type)
        vb = scale_timedelta(context, builder, vb, tb, sig.return_type)
        builder.store(builder.add(va, vb), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.sub, *TIMEDELTA_BINOP_SIG)
@lower_builtin(operator.isub, *TIMEDELTA_BINOP_SIG)
def timedelta_sub_impl(context, builder, sig, args):
    [va, vb] = args
    [ta, tb] = sig.args
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = scale_timedelta(context, builder, va, ta, sig.return_type)
        vb = scale_timedelta(context, builder, vb, tb, sig.return_type)
        builder.store(builder.sub(va, vb), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


def _timedelta_times_number(context, builder, td_arg, td_type,
                            number_arg, number_type, return_type):
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, is_not_nat(builder, td_arg)):
        if isinstance(number_type, types.Float):
            val = builder.sitofp(td_arg, number_arg.type)
            val = builder.fmul(val, number_arg)
            val = _cast_to_timedelta(context, builder, val)
        else:
            val = builder.mul(td_arg, number_arg)
        # The scaling is required for ufunc np.multiply() with an explicit
        # output in a different unit.
        val = scale_timedelta(context, builder, val, td_type, return_type)
        builder.store(val, ret)
    return builder.load(ret)


@lower_builtin(operator.mul, types.NPTimedelta, types.Integer)
@lower_builtin(operator.imul, types.NPTimedelta, types.Integer)
@lower_builtin(operator.mul, types.NPTimedelta, types.Float)
@lower_builtin(operator.imul, types.NPTimedelta, types.Float)
def timedelta_times_number(context, builder, sig, args):
    res = _timedelta_times_number(context, builder,
                                  args[0], sig.args[0], args[1], sig.args[1],
                                  sig.return_type)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.mul, types.Integer, types.NPTimedelta)
@lower_builtin(operator.imul, types.Integer, types.NPTimedelta)
@lower_builtin(operator.mul, types.Float, types.NPTimedelta)
@lower_builtin(operator.imul, types.Float, types.NPTimedelta)
def number_times_timedelta(context, builder, sig, args):
    res = _timedelta_times_number(context, builder,
                                  args[1], sig.args[1], args[0], sig.args[0],
                                  sig.return_type)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.truediv, types.NPTimedelta, types.Integer)
@lower_builtin(operator.itruediv, types.NPTimedelta, types.Integer)
@lower_builtin(operator.floordiv, types.NPTimedelta, types.Integer)
@lower_builtin(operator.ifloordiv, types.NPTimedelta, types.Integer)
@lower_builtin(operator.truediv, types.NPTimedelta, types.Float)
@lower_builtin(operator.itruediv, types.NPTimedelta, types.Float)
@lower_builtin(operator.floordiv, types.NPTimedelta, types.Float)
@lower_builtin(operator.ifloordiv, types.NPTimedelta, types.Float)
def timedelta_over_number(context, builder, sig, args):
    td_arg, number_arg = args
    number_type = sig.args[1]
    ret = alloc_timedelta_result(builder)
    ok = builder.and_(is_not_nat(builder, td_arg),
                      builder.not_(cgutils.is_scalar_zero_or_nan(builder, number_arg)))
    with cgutils.if_likely(builder, ok):
        # Denominator is non-zero, non-NaN
        if isinstance(number_type, types.Float):
            val = builder.sitofp(td_arg, number_arg.type)
            val = builder.fdiv(val, number_arg)
            val = _cast_to_timedelta(context, builder, val)
        else:
            val = builder.sdiv(td_arg, number_arg)
        # The scaling is required for ufuncs np.*divide() with an explicit
        # output in a different unit.
        val = scale_timedelta(context, builder, val,
                              sig.args[0], sig.return_type)
        builder.store(val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.truediv, *TIMEDELTA_BINOP_SIG)
@lower_builtin(operator.itruediv, *TIMEDELTA_BINOP_SIG)
def timedelta_over_timedelta(context, builder, sig, args):
    [va, vb] = args
    [ta, tb] = sig.args
    not_nan = are_not_nat(builder, [va, vb])
    ll_ret_type = context.get_value_type(sig.return_type)
    ret = cgutils.alloca_once(builder, ll_ret_type, name='ret')
    builder.store(Constant.real(ll_ret_type, float('nan')), ret)
    with cgutils.if_likely(builder, not_nan):
        va, vb = normalize_timedeltas(context, builder, va, vb, ta, tb)
        va = builder.sitofp(va, ll_ret_type)
        vb = builder.sitofp(vb, ll_ret_type)
        builder.store(builder.fdiv(va, vb), ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)


if numpy_support.numpy_version >= (1, 16):
    # np 1.16 added support for:
    # * np.floor_divide on mm->q
    # * np.remainder on mm->m

    @lower_builtin(operator.floordiv, *TIMEDELTA_BINOP_SIG)
    def timedelta_floor_div_timedelta(context, builder, sig, args):
        [va, vb] = args
        [ta, tb] = sig.args
        ll_ret_type = context.get_value_type(sig.return_type)
        not_nan = are_not_nat(builder, [va, vb])
        ret = cgutils.alloca_once(builder, ll_ret_type, name='ret')
        zero = Constant.int(ll_ret_type, 0)
        one = Constant.int(ll_ret_type, 1)
        builder.store(zero, ret)
        with cgutils.if_likely(builder, not_nan):
            va, vb = normalize_timedeltas(context, builder, va, vb, ta, tb)
            # is the denominator zero or NaT?
            denom_ok = builder.not_(builder.icmp_signed('==', vb, zero))
            with cgutils.if_likely(builder, denom_ok):
                # is either arg negative?
                vaneg = builder.icmp_signed('<', va, zero)
                neg = builder.or_(vaneg, builder.icmp_signed('<', vb, zero))
                with builder.if_else(neg) as (then, otherwise):
                    with then:  # one or more value negative
                        with builder.if_else(vaneg) as (negthen, negotherwise):
                            with negthen:
                                top = builder.sub(va, one)
                                div = builder.sdiv(top, vb)
                                builder.store(div, ret)
                            with negotherwise:
                                top = builder.add(va, one)
                                div = builder.sdiv(top, vb)
                                builder.store(div, ret)
                    with otherwise:
                        div = builder.sdiv(va, vb)
                        builder.store(div, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    def timedelta_mod_timedelta(context, builder, sig, args):
        # inspired by https://github.com/numpy/numpy/blob/fe8072a12d65e43bd2e0b0f9ad67ab0108cc54b3/numpy/core/src/umath/loops.c.src#L1424
        # alg is basically as `a % b`:
        # if a or b is NaT return NaT
        # elseif b is 0 return NaT
        # else pretend a and b are int and do pythonic int modulus

        [va, vb] = args
        [ta, tb] = sig.args
        not_nan = are_not_nat(builder, [va, vb])
        ll_ret_type = context.get_value_type(sig.return_type)
        ret = alloc_timedelta_result(builder)
        builder.store(NAT, ret)
        zero = Constant.int(ll_ret_type, 0)
        with cgutils.if_likely(builder, not_nan):
            va, vb = normalize_timedeltas(context, builder, va, vb, ta, tb)
            # is the denominator zero or NaT?
            denom_ok = builder.not_(builder.icmp_signed('==', vb, zero))
            with cgutils.if_likely(builder, denom_ok):
                # is either arg negative?
                vapos = builder.icmp_signed('>', va, zero)
                vbpos = builder.icmp_signed('>', vb, zero)
                rem = builder.srem(va, vb)
                cond = builder.or_(builder.and_(vapos, vbpos),
                                   builder.icmp_signed('==', rem, zero))
                with builder.if_else(cond) as (then, otherwise):
                    with then:
                        builder.store(rem, ret)
                    with otherwise:
                        builder.store(builder.add(rem, vb), ret)

        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

# Comparison operators on timedelta64


def _create_timedelta_comparison_impl(ll_op, default_value):
    def impl(context, builder, sig, args):
        [va, vb] = args
        [ta, tb] = sig.args
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (then, otherwise):
            with then:
                try:
                    norm_a, norm_b = normalize_timedeltas(
                        context, builder, va, vb, ta, tb)
                except RuntimeError:
                    # Cannot normalize units => the values are unequal (except if NaT)
                    builder.store(default_value, ret)
                else:
                    builder.store(builder.icmp(ll_op, norm_a, norm_b), ret)
            with otherwise:
                if numpy_support.numpy_version < (1, 16):
                    # No scaling when comparing NaTs
                    builder.store(builder.icmp(ll_op, va, vb), ret)
                else:
                    # NumPy >= 1.16 switched to NaT ==/>=/>/</<= NaT being
                    # False and NaT != <anything, including NaT> being True
                    if ll_op == lc.ICMP_NE:
                        builder.store(cgutils.true_bit, ret)
                    else:
                        builder.store(cgutils.false_bit, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


def _create_timedelta_ordering_impl(ll_op):
    def impl(context, builder, sig, args):
        [va, vb] = args
        [ta, tb] = sig.args
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (then, otherwise):
            with then:
                norm_a, norm_b = normalize_timedeltas(
                    context, builder, va, vb, ta, tb)
                builder.store(builder.icmp(ll_op, norm_a, norm_b), ret)
            with otherwise:
                if numpy_support.numpy_version < (1, 16):
                    # No scaling when comparing NaT with something else
                    # (i.e. NaT is <= everything else, since it's the smallest
                    #  int64 value)
                    builder.store(builder.icmp(ll_op, va, vb), ret)
                else:
                    # NumPy >= 1.16 switched to NaT >=/>/</<= NaT being False
                    builder.store(cgutils.false_bit, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


timedelta_eq_timedelta_impl = _create_timedelta_comparison_impl(
    lc.ICMP_EQ, cgutils.false_bit)
timedelta_ne_timedelta_impl = _create_timedelta_comparison_impl(
    lc.ICMP_NE, cgutils.true_bit)
timedelta_lt_timedelta_impl = _create_timedelta_ordering_impl(lc.ICMP_SLT)
timedelta_le_timedelta_impl = _create_timedelta_ordering_impl(lc.ICMP_SLE)
timedelta_gt_timedelta_impl = _create_timedelta_ordering_impl(lc.ICMP_SGT)
timedelta_ge_timedelta_impl = _create_timedelta_ordering_impl(lc.ICMP_SGE)

for op_, func in [(operator.eq, timedelta_eq_timedelta_impl),
                  (operator.ne, timedelta_ne_timedelta_impl),
                  (operator.lt,  timedelta_lt_timedelta_impl),
                  (operator.le, timedelta_le_timedelta_impl),
                  (operator.gt,  timedelta_gt_timedelta_impl),
                  (operator.ge, timedelta_ge_timedelta_impl)]:
    lower_builtin(op_, *TIMEDELTA_BINOP_SIG)(func)


# Arithmetic on datetime64

def is_leap_year(builder, year_val):
    """
    Return a predicate indicating whether *year_val* (offset by 1970) is a
    leap year.
    """
    actual_year = builder.add(year_val, Constant.int(DATETIME64, 1970))
    multiple_of_4 = cgutils.is_null(
        builder, builder.and_(actual_year, Constant.int(DATETIME64, 3)))
    not_multiple_of_100 = cgutils.is_not_null(
        builder, builder.srem(actual_year, Constant.int(DATETIME64, 100)))
    multiple_of_400 = cgutils.is_null(
        builder, builder.srem(actual_year, Constant.int(DATETIME64, 400)))
    return builder.and_(multiple_of_4,
                        builder.or_(not_multiple_of_100, multiple_of_400))


def year_to_days(builder, year_val):
    """
    Given a year *year_val* (offset to 1970), return the number of days
    since the 1970 epoch.
    """
    # The algorithm below is copied from Numpy's get_datetimestruct_days()
    # (src/multiarray/datetime.c)
    ret = cgutils.alloca_once(builder, TIMEDELTA64)
    # First approximation
    days = scale_by_constant(builder, year_val, 365)
    # Adjust for leap years
    with builder.if_else(cgutils.is_neg_int(builder, year_val)) \
            as (if_neg, if_pos):
        with if_pos:
            # At or after 1970:
            # 1968 is the closest leap year before 1970.
            # Exclude the current year, so add 1.
            from_1968 = add_constant(builder, year_val, 1)
            # Add one day for each 4 years
            p_days = builder.add(days,
                                 unscale_by_constant(builder, from_1968, 4))
            # 1900 is the closest previous year divisible by 100
            from_1900 = add_constant(builder, from_1968, 68)
            # Subtract one day for each 100 years
            p_days = builder.sub(p_days,
                                 unscale_by_constant(builder, from_1900, 100))
            # 1600 is the closest previous year divisible by 400
            from_1600 = add_constant(builder, from_1900, 300)
            # Add one day for each 400 years
            p_days = builder.add(p_days,
                                 unscale_by_constant(builder, from_1600, 400))
            builder.store(p_days, ret)
        with if_neg:
            # Before 1970:
            # NOTE `year_val` is negative, and so will be `from_1972` and `from_2000`.
            # 1972 is the closest later year after 1970.
            # Include the current year, so subtract 2.
            from_1972 = add_constant(builder, year_val, -2)
            # Subtract one day for each 4 years (`from_1972` is negative)
            n_days = builder.add(days,
                                 unscale_by_constant(builder, from_1972, 4))
            # 2000 is the closest later year divisible by 100
            from_2000 = add_constant(builder, from_1972, -28)
            # Add one day for each 100 years
            n_days = builder.sub(n_days,
                                 unscale_by_constant(builder, from_2000, 100))
            # 2000 is also the closest later year divisible by 400
            # Subtract one day for each 400 years
            n_days = builder.add(n_days,
                                 unscale_by_constant(builder, from_2000, 400))
            builder.store(n_days, ret)
    return builder.load(ret)


def reduce_datetime_for_unit(builder, dt_val, src_unit, dest_unit):
    dest_unit_code = npdatetime_helpers.DATETIME_UNITS[dest_unit]
    src_unit_code = npdatetime_helpers.DATETIME_UNITS[src_unit]
    if dest_unit_code < 2 or src_unit_code >= 2:
        return dt_val, src_unit
    # Need to compute the day ordinal for *dt_val*
    if src_unit_code == 0:
        # Years to days
        year_val = dt_val
        days_val = year_to_days(builder, year_val)

    else:
        # Months to days
        leap_array = cgutils.global_constant(builder, "leap_year_months_acc",
                                             leap_year_months_acc)
        normal_array = cgutils.global_constant(builder, "normal_year_months_acc",
                                               normal_year_months_acc)

        days = cgutils.alloca_once(builder, TIMEDELTA64)

        # First compute year number and month number
        year, month = cgutils.divmod_by_constant(builder, dt_val, 12)

        # Then deduce the number of days
        with builder.if_else(is_leap_year(builder, year)) as (then, otherwise):
            with then:
                addend = builder.load(cgutils.gep(builder, leap_array,
                                                  0, month, inbounds=True))
                builder.store(addend, days)
            with otherwise:
                addend = builder.load(cgutils.gep(builder, normal_array,
                                                  0, month, inbounds=True))
                builder.store(addend, days)

        days_val = year_to_days(builder, year)
        days_val = builder.add(days_val, builder.load(days))

    if dest_unit_code == 2:
        # Need to scale back to weeks
        weeks, _ = cgutils.divmod_by_constant(builder, days_val, 7)
        return weeks, 'W'
    else:
        return days_val, 'D'


def convert_datetime_for_arith(builder, dt_val, src_unit, dest_unit):
    """
    Convert datetime *dt_val* from *src_unit* to *dest_unit*.
    """
    # First partial conversion to days or weeks, if necessary.
    dt_val, dt_unit = reduce_datetime_for_unit(
        builder, dt_val, src_unit, dest_unit)
    # Then multiply by the remaining constant factor.
    dt_factor = npdatetime_helpers.get_timedelta_conversion_factor(dt_unit, dest_unit)
    if dt_factor is None:
        # This can happen when using explicit output in a ufunc.
        raise NotImplementedError("cannot convert datetime64 from %r to %r"
                                  % (src_unit, dest_unit))
    return scale_by_constant(builder, dt_val, dt_factor)


def _datetime_timedelta_arith(ll_op_name):
    def impl(context, builder, dt_arg, dt_unit,
             td_arg, td_unit, ret_unit):
        ret = alloc_timedelta_result(builder)
        with cgutils.if_likely(builder, are_not_nat(builder, [dt_arg, td_arg])):
            dt_arg = convert_datetime_for_arith(builder, dt_arg,
                                                dt_unit, ret_unit)
            td_factor = npdatetime_helpers.get_timedelta_conversion_factor(
                td_unit, ret_unit)
            td_arg = scale_by_constant(builder, td_arg, td_factor)
            ret_val = getattr(builder, ll_op_name)(dt_arg, td_arg)
            builder.store(ret_val, ret)
        return builder.load(ret)
    return impl


_datetime_plus_timedelta = _datetime_timedelta_arith('add')
_datetime_minus_timedelta = _datetime_timedelta_arith('sub')

# datetime64 + timedelta64


@lower_builtin(operator.add, types.NPDatetime, types.NPTimedelta)
@lower_builtin(operator.iadd, types.NPDatetime, types.NPTimedelta)
def datetime_plus_timedelta(context, builder, sig, args):
    dt_arg, td_arg = args
    dt_type, td_type = sig.args
    res = _datetime_plus_timedelta(context, builder,
                                   dt_arg, dt_type.unit,
                                   td_arg, td_type.unit,
                                   sig.return_type.unit)
    return impl_ret_untracked(context, builder, sig.return_type, res)


@lower_builtin(operator.add, types.NPTimedelta, types.NPDatetime)
@lower_builtin(operator.iadd, types.NPTimedelta, types.NPDatetime)
def timedelta_plus_datetime(context, builder, sig, args):
    td_arg, dt_arg = args
    td_type, dt_type = sig.args
    res = _datetime_plus_timedelta(context, builder,
                                   dt_arg, dt_type.unit,
                                   td_arg, td_type.unit,
                                   sig.return_type.unit)
    return impl_ret_untracked(context, builder, sig.return_type, res)

# datetime64 - timedelta64


@lower_builtin(operator.sub, types.NPDatetime, types.NPTimedelta)
@lower_builtin(operator.isub, types.NPDatetime, types.NPTimedelta)
def datetime_minus_timedelta(context, builder, sig, args):
    dt_arg, td_arg = args
    dt_type, td_type = sig.args
    res = _datetime_minus_timedelta(context, builder,
                                    dt_arg, dt_type.unit,
                                    td_arg, td_type.unit,
                                    sig.return_type.unit)
    return impl_ret_untracked(context, builder, sig.return_type, res)

# datetime64 - datetime64


@lower_builtin(operator.sub, types.NPDatetime, types.NPDatetime)
def datetime_minus_datetime(context, builder, sig, args):
    va, vb = args
    ta, tb = sig.args
    unit_a = ta.unit
    unit_b = tb.unit
    ret_unit = sig.return_type.unit
    ret = alloc_timedelta_result(builder)
    with cgutils.if_likely(builder, are_not_nat(builder, [va, vb])):
        va = convert_datetime_for_arith(builder, va, unit_a, ret_unit)
        vb = convert_datetime_for_arith(builder, vb, unit_b, ret_unit)
        ret_val = builder.sub(va, vb)
        builder.store(ret_val, ret)
    res = builder.load(ret)
    return impl_ret_untracked(context, builder, sig.return_type, res)

# datetime64 comparisons


def _create_datetime_comparison_impl(ll_op):
    def impl(context, builder, sig, args):
        va, vb = args
        ta, tb = sig.args
        unit_a = ta.unit
        unit_b = tb.unit
        ret_unit = npdatetime_helpers.get_best_unit(unit_a, unit_b)
        ret = alloc_boolean_result(builder)
        with builder.if_else(are_not_nat(builder, [va, vb])) as (then, otherwise):
            with then:
                norm_a = convert_datetime_for_arith(
                    builder, va, unit_a, ret_unit)
                norm_b = convert_datetime_for_arith(
                    builder, vb, unit_b, ret_unit)
                ret_val = builder.icmp(ll_op, norm_a, norm_b)
                builder.store(ret_val, ret)
            with otherwise:
                if numpy_support.numpy_version < (1, 16):
                    # No scaling when comparing NaTs
                    ret_val = builder.icmp(ll_op, va, vb)
                else:
                    if ll_op == lc.ICMP_NE:
                        ret_val = cgutils.true_bit
                    else:
                        ret_val = cgutils.false_bit
                builder.store(ret_val, ret)
        res = builder.load(ret)
        return impl_ret_untracked(context, builder, sig.return_type, res)

    return impl


datetime_eq_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_EQ)
datetime_ne_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_NE)
datetime_lt_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_SLT)
datetime_le_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_SLE)
datetime_gt_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_SGT)
datetime_ge_datetime_impl = _create_datetime_comparison_impl(lc.ICMP_SGE)

for op, func in [(operator.eq, datetime_eq_datetime_impl),
                 (operator.ne, datetime_ne_datetime_impl),
                 (operator.lt, datetime_lt_datetime_impl),
                 (operator.le, datetime_le_datetime_impl),
                 (operator.gt, datetime_gt_datetime_impl),
                 (operator.ge, datetime_ge_datetime_impl)]:
    lower_builtin(op, *[types.NPDatetime]*2)(func)


########################################################################
# datetime/timedelta fmax/fmin maximum/minimum support

def datetime_max_impl(context, builder, sig, args):
    # just a regular int64 max avoiding nats.
    # note this could be optimizing relying on the actual value of NAT
    # but as NumPy doesn't rely on this, this seems more resilient
    in1, in2 = args
    in1_not_nat = is_not_nat(builder, in1)
    in2_not_nat = is_not_nat(builder, in2)
    in1_ge_in2 = builder.icmp(lc.ICMP_SGE, in1, in2)
    res = builder.select(in1_ge_in2, in1, in2)
    res = builder.select(in1_not_nat, res, in2)
    res = builder.select(in2_not_nat, res, in1)

    return impl_ret_untracked(context, builder, sig.return_type, res)


def datetime_min_impl(context, builder, sig, args):
    # just a regular int64 min avoiding nats.
    # note this could be optimizing relying on the actual value of NAT
    # but as NumPy doesn't rely on this, this seems more resilient
    in1, in2 = args
    in1_not_nat = is_not_nat(builder, in1)
    in2_not_nat = is_not_nat(builder, in2)
    in1_le_in2 = builder.icmp(lc.ICMP_SLE, in1, in2)
    res = builder.select(in1_le_in2, in1, in2)
    res = builder.select(in1_not_nat, res, in2)
    res = builder.select(in2_not_nat, res, in1)

    return impl_ret_untracked(context, builder, sig.return_type, res)


def timedelta_max_impl(context, builder, sig, args):
    # just a regular int64 max avoiding nats.
    # note this could be optimizing relying on the actual value of NAT
    # but as NumPy doesn't rely on this, this seems more resilient
    in1, in2 = args
    in1_not_nat = is_not_nat(builder, in1)
    in2_not_nat = is_not_nat(builder, in2)
    in1_ge_in2 = builder.icmp(lc.ICMP_SGE, in1, in2)
    res = builder.select(in1_ge_in2, in1, in2)
    res = builder.select(in1_not_nat, res, in2)
    res = builder.select(in2_not_nat, res, in1)

    return impl_ret_untracked(context, builder, sig.return_type, res)


def timedelta_min_impl(context, builder, sig, args):
    # just a regular int64 min avoiding nats.
    # note this could be optimizing relying on the actual value of NAT
    # but as NumPy doesn't rely on this, this seems more resilient
    in1, in2 = args
    in1_not_nat = is_not_nat(builder, in1)
    in2_not_nat = is_not_nat(builder, in2)
    in1_le_in2 = builder.icmp(lc.ICMP_SLE, in1, in2)
    res = builder.select(in1_le_in2, in1, in2)
    res = builder.select(in1_not_nat, res, in2)
    res = builder.select(in2_not_nat, res, in1)

    return impl_ret_untracked(context, builder, sig.return_type, res)


def _cast_to_timedelta(context, builder, val):
    temp = builder.alloca(TIMEDELTA64)
    val_is_nan = builder.fcmp_unordered('uno', val, val)
    with builder.if_else(val_is_nan) as (
            then, els):
        with then:
            # NaN does not guarantee to cast to NAT.
            # We should store NAT explicitly.
            builder.store(NAT, temp)
        with els:
            builder.store(builder.fptosi(val, TIMEDELTA64), temp)
    return builder.load(temp)


@lower_builtin(np.isnat, types.NPDatetime)
@lower_builtin(np.isnat, types.NPTimedelta)
def _np_isnat_impl(context, builder, sig, args):
    return npyfuncs.np_datetime_isnat_impl(context, builder, sig, args)



import numba.testing
from numba.core import config

if config.ENABLE_CUDASIM:
    from .simulator_init import *
else:
    from .device_init import *
    from .device_init import _auto_device


def test(*args, **kwargs):
    if not is_available():
        raise cuda_error()

    return numba.testing.test("numba.cuda.tests", *args, **kwargs)



import math
import sys

import numpy as np

from numba.core.compiler import compile_isolated
import unittest


class TestAutoConstants(unittest.TestCase):
    def test_numpy_nan(self):
        def pyfunc():
            return np.nan

        cres = compile_isolated(pyfunc, ())
        cfunc = cres.entry_point

        self.assertTrue(math.isnan(pyfunc()))
        self.assertTrue(math.isnan(cfunc()))

    def test_sys_constant(self):
        def pyfunc():
            return sys.hexversion

        cres = compile_isolated(pyfunc, ())
        cfunc = cres.entry_point

        self.assertEqual(pyfunc(), cfunc())


if __name__ == '__main__':
    unittest.main()




import tempfile
import os
import numpy as np

import unittest
from numba import roc
from numba.core import types
from numba.roc import compiler
from numba.roc.hsadrv.driver import hsa as hsart
from numba.roc.hsadrv.driver import BrigModule, Executable, Program


def copy_kernel(out, inp):
    out[0] = inp[0]


def copy_kernel_1d(out, inp):
    i = roc.get_global_id(0)
    if i < out.size:
        out[i] = inp[i]


def assign_value(out, inp):
    i = roc.get_global_id(0)
    if i < out.size:
        out[i] = inp


class TestCodeGeneration(unittest.TestCase):
    def test_copy_kernel(self):
        arytype = types.float32[:]
        kernel = compiler.compile_kernel(copy_kernel, [arytype] * 2)
        self.assertIn(".globl\t{0}".format(kernel.entry_name),
                      kernel.assembly)

    def test_copy_kernel_1d(self):
        arytype = types.float32[:]
        kernel = compiler.compile_kernel(copy_kernel_1d, [arytype] * 2)
        self.assertIn(".globl\t{0}".format(kernel.entry_name),
                      kernel.assembly)


class _TestBase(unittest.TestCase):
    def setUp(self):
        self.gpu = [a for a in hsart.agents if a.is_component][0]
        self.cpu = [a for a in hsart.agents if not a.is_component][0]
        self.queue = self.gpu.create_queue_multi(self.gpu.queue_max_size)

    def tearDown(self):
        del self.queue
        del self.gpu
        del self.cpu


class TestExecution(unittest.TestCase):
    def test_hsa_kernel(self):
        src = np.arange(1024, dtype=np.float32)
        dst = np.zeros_like(src)

        # Compiler kernel
        arytype = types.float32[::1]
        kernel = compiler.compile_kernel(copy_kernel_1d, [arytype] * 2)

        # Run kernel
        kernel[src.size // 256, 256](dst, src)

        np.testing.assert_equal(src, dst)


class TestKernelArgument(unittest.TestCase):
    def _test_template(self, nbtype, src):
        dtype = np.dtype(str(nbtype))
        dst = np.zeros(1, dtype=dtype)
        src = dtype.type(src)
        arytype = nbtype[::1]
        kernel = compiler.compile_kernel(assign_value, [arytype, nbtype])
        kernel[1, 1](dst, src)
        self.assertEqual(dst[0], src)

    def test_float64(self):
        self._test_template(nbtype=types.float64, src=1. / 3.)

    def test_float32(self):
        self._test_template(nbtype=types.float32, src=1. / 3.)

    def test_int32(self):
        self._test_template(nbtype=types.int32, src=123)

    def test_int16(self):
        self._test_template(nbtype=types.int16, src=123)

    def test_complex64(self):
        self._test_template(nbtype=types.complex64, src=12 + 34j)

    def test_complex128(self):
        self._test_template(nbtype=types.complex128, src=12 + 34j)


def udt_devfunc(a, i):
    return a[i]


class TestDeviceFunction(unittest.TestCase):
    def test_device_function(self):
        src = np.arange(10, dtype=np.int32)
        dst = np.zeros_like(src)

        arytype = types.int32[::1]
        devfn = compiler.compile_device(udt_devfunc, arytype.dtype,
                                        [arytype, types.intp])

        def udt_devfunc_caller(dst, src):
            i = roc.get_global_id(0)
            if i < dst.size:
                dst[i] = devfn(src, i)

        kernel = compiler.compile_kernel(udt_devfunc_caller,
                                         [arytype, arytype])

        kernel[src.size, 1](dst, src)
        np.testing.assert_equal(dst, src)


if __name__ == '__main__':
    unittest.main()

